from unittest import mock

import pytest

from bec_lib.scans import ScanObject


@pytest.fixture
def scan_obj(bec_client_mock):
    scan_info = {
        "class": "FermatSpiralScan",
        "arg_input": {"device": "device", "start": "float", "stop": "float"},
        "required_kwargs": ["step", "relative"],
        "arg_bundle_size": {"bundle": 3, "min": 2, "max": 2},
        "scan_report_hint": "table",
        "doc": (
            "\n        A scan following Fermat's spiral.\n\n        Args:\n            *args: pairs"
            " of device / start position / end position / steps arguments\n            relative:"
            " Start from an absolute or relative position\n            burst: number of acquisition"
            " per point\n            optim_trajectory: routine used for the trajectory"
            " optimization, e.g. 'corridor'. Default: None\n\n        Returns:\n\n       "
            " Examples:\n            >>> scans.fermat_scan(dev.motor1, -5, 5, dev.motor2, -5, 5,"
            ' step=0.5, exp_time=0.1, relative=True, optim_trajectory="corridor")\n\n        '
        ),
    }
    scan_name = "fermat_scan"
    obj = ScanObject(scan_name, scan_info, bec_client_mock)
    with mock.patch.object(bec_client_mock, "alarm_handler"):
        yield obj


@pytest.fixture
def dev(scan_obj):
    devices = scan_obj.client.device_manager.devices
    yield devices


def test_scan_object_raises(scan_obj):
    with pytest.raises(TypeError):
        scan_obj._run()


def test_scan_object_raises_not_enough_bundles(scan_obj, dev):
    with pytest.raises(TypeError):
        scan_obj._run(dev.samx, -5, 5, step=0.5, exp_time=0.1, relative=False)


def test_scan_object_raises_too_many_bundles(scan_obj, dev):
    with pytest.raises(TypeError):
        scan_obj._run(
            dev.samx,
            -5,
            5,
            dev.samy,
            -5,
            5,
            dev.samz,
            -5,
            5,
            step=0.5,
            exp_time=0.1,
            relative=False,
        )


def test_scan_object(scan_obj, dev):
    with mock.patch("bec_lib.scan_report.ScanReport.from_request") as report:
        scan_obj._run(dev.samx, -5, 5, dev.samy, -5, 5, step=0.5, exp_time=0.1, relative=False)
        report().wait.assert_not_called()


def test_scan_object_wo_live_updates(scan_obj, dev):
    scan_obj.client.live_updates = None
    with mock.patch("bec_lib.scan_report.ScanReport.from_request") as report:
        scan_obj._run(dev.samx, -5, 5, dev.samy, -5, 5, step=0.5, exp_time=0.1, relative=False)
        report().wait.assert_not_called()


def test_scan_object_file_suffix(scan_obj, dev):
    with mock.patch("bec_lib.scan_report.ScanReport.from_request") as scan_report:
        with mock.patch.object(scan_obj.client, "get_global_var", return_value="test_sample"):
            scan_obj._run(
                dev.samx,
                -5,
                5,
                dev.samy,
                -5,
                5,
                step=0.5,
                exp_time=0.1,
                relative=False,
                file_suffix="testsample",
            )
            assert (
                scan_report.call_args.args[0].metadata["file_writer_data"]["file_suffix"]
                == "testsample"
            )


@pytest.mark.parametrize(
    "file_suffix, file_suffix_raises",
    [
        ("testsample", False),
        ("testsample_", False),
        ("testsample-", False),
        ("ötestsample", True),
        ("123sample", False),
        ("../sample", True),
        ("sample/123", True),
        ("sample\\123", True),
        ("sample123\\", True),
        ("sample123/", True),
        ("sample123.", True),
    ],
)
def test_scan_object_raises_on_non_ascii_chars(scan_obj, dev, file_suffix, file_suffix_raises):
    with mock.patch("bec_lib.scan_report.ScanReport.from_request") as scan_report:
        with mock.patch.object(scan_obj.client, "get_global_var", return_value="test_sample"):
            if file_suffix_raises:
                with pytest.raises(ValueError):
                    scan_obj._run(
                        dev.samx,
                        -5,
                        5,
                        dev.samy,
                        -5,
                        5,
                        step=0.5,
                        exp_time=0.1,
                        relative=False,
                        file_suffix=file_suffix,
                    )
            else:
                scan_obj._run(
                    dev.samx,
                    -5,
                    5,
                    dev.samy,
                    -5,
                    5,
                    step=0.5,
                    exp_time=0.1,
                    relative=False,
                    file_suffix=file_suffix,
                )
                assert (
                    scan_report.call_args.args[0].metadata["file_writer_data"]["file_suffix"]
                    == file_suffix
                )


@pytest.mark.parametrize(
    "file_dir, file_suffix_raises",
    [
        ("/tmp/data/", False),
        ("/tmp/data./", True),
        ("/tmp/data-1/", False),
        ("/tmp/data_1/", False),
        ("sample\\123", True),
        ("//sample//123//", False),
    ],
)
def test_scan_object_raises_on_non_ascii_chars_dir(scan_obj, dev, file_dir, file_suffix_raises):
    with mock.patch("bec_lib.scan_report.ScanReport.from_request") as scan_report:
        with mock.patch.object(scan_obj.client, "get_global_var", return_value="test_sample"):
            if file_suffix_raises:
                with pytest.raises(ValueError):
                    scan_obj._run(
                        dev.samx,
                        -5,
                        5,
                        dev.samy,
                        -5,
                        5,
                        step=0.5,
                        exp_time=0.1,
                        relative=False,
                        file_directory=file_dir,
                    )
            else:
                scan_obj._run(
                    dev.samx,
                    -5,
                    5,
                    dev.samy,
                    -5,
                    5,
                    step=0.5,
                    exp_time=0.1,
                    relative=False,
                    file_directory=file_dir,
                )
                assert scan_report.call_args.args[0].metadata["file_writer_data"][
                    "file_directory"
                ] == file_dir.strip("/")


def test_scan_object_receives_sample_name(scan_obj, dev):
    with mock.patch("bec_lib.scan_report.ScanReport.from_request") as scan_report:
        with mock.patch.object(scan_obj.client, "get_global_var", return_value="test_sample"):
            scan_obj._run(dev.samx, -5, 5, dev.samy, -5, 5, step=0.5, exp_time=0.1, relative=False)
            assert scan_report.call_args.args[0].metadata["sample_name"] == "test_sample"


def test_scan_object_receives_scan_group(scan_obj, dev):
    scan_obj.client.scans._scan_group = "group_id"
    with mock.patch.object(scan_obj.client, "alarm_handler"):
        with mock.patch("bec_lib.scan_report.ScanReport.from_request") as scan_report:
            with mock.patch.object(scan_obj.client, "get_global_var", return_value="test_sample"):
                scan_obj._run(
                    dev.samx, -5, 5, dev.samy, -5, 5, step=0.5, exp_time=0.1, relative=False
                )
                assert scan_report.call_args.args[0].metadata["queue_group"] == "group_id"


def test_scan_object_receives_scan_def_id(scan_obj, dev):
    scan_obj.client.scans._scan_def_id = "scan_def_id"
    with mock.patch.object(scan_obj.client, "alarm_handler"):
        with mock.patch("bec_lib.scan_report.ScanReport.from_request") as scan_report:
            with mock.patch.object(scan_obj.client, "get_global_var", return_value="test_sample"):
                scan_obj._run(
                    dev.samx, -5, 5, dev.samy, -5, 5, step=0.5, exp_time=0.1, relative=False
                )
                assert scan_report.call_args.args[0].metadata["scan_def_id"] == "scan_def_id"


def test_scan_object_receives_dataset_id_on_hold(scan_obj, dev):
    scan_obj.client.scans._dataset_id_on_hold = "dataset_id_on_hold"
    with mock.patch.object(scan_obj.client, "alarm_handler"):
        with mock.patch("bec_lib.scan_report.ScanReport.from_request") as scan_report:
            with mock.patch.object(scan_obj.client, "get_global_var", return_value="test_sample"):
                scan_obj._run(
                    dev.samx, -5, 5, dev.samy, -5, 5, step=0.5, exp_time=0.1, relative=False
                )
                assert (
                    scan_report.call_args.args[0].metadata["dataset_id_on_hold"]
                    == "dataset_id_on_hold"
                )



from unittest import mock

import numpy as np
import pytest

from bec_lib import messages
from bec_lib.async_data import AsyncDataHandler
from bec_lib.endpoints import MessageEndpoints

# pylint: disable=protected-access
# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=redefined-outer-name


@pytest.fixture
def async_data():
    producer = mock.MagicMock()
    yield AsyncDataHandler(producer)


def test_process_async_data_replace(async_data):
    data = [
        {
            "data": messages.DeviceMessage(
                signals={"data": {"value": np.zeros((10, 10))}},
                metadata={"async_update": "replace"},
            )
        }
        for ii in range(10)
    ]
    res = async_data.process_async_data(data)
    assert res["data"]["value"].shape == (10, 10)


def test_process_async_multiple_signals(async_data):
    data = [
        {
            "data": messages.DeviceMessage(
                signals={
                    "signal1": {"value": np.zeros((10, 10))},
                    "signal2": {"value": np.zeros((20, 20))},
                },
                metadata={"async_update": "replace"},
            )
        }
        for ii in range(10)
    ]
    res = async_data.process_async_data(data)
    assert res["signal1"]["value"].shape == (10, 10)
    assert res["signal2"]["value"].shape == (20, 20)


def test_process_async_data_extend(async_data):
    data = [
        {
            "data": messages.DeviceMessage(
                signals={"data": {"value": np.zeros((10, 10))}}, metadata={"async_update": "extend"}
            )
        }
        for ii in range(10)
    ]
    res = async_data.process_async_data(data)
    assert res["data"]["value"].shape == (100, 10)


def test_process_async_update_append(async_data):
    data = [
        {
            "data": messages.DeviceMessage(
                signals={"data": {"value": np.zeros((10, 10))}}, metadata={"async_update": "append"}
            )
        }
        for ii in range(10)
    ]
    res = async_data.process_async_data(data)
    assert res["data"][0]["value"].shape == (10, 10)
    assert len(res["data"]) == 10


def test_process_async_data_single(async_data):
    data = [
        {
            "data": messages.DeviceMessage(
                signals={"data": {"value": np.zeros((10, 10))}}, metadata={}
            )
        }
    ]
    res = async_data.process_async_data(data)
    assert res["data"]["value"].shape == (10, 10)


def test_get_async_data_for_scan():
    producer = mock.MagicMock()
    async_data = AsyncDataHandler(producer)
    producer.keys.return_value = [
        MessageEndpoints.device_async_readback("scan_id", "samx").endpoint.encode(),
        MessageEndpoints.device_async_readback("scan_id", "samy").endpoint.encode(),
    ]
    with mock.patch.object(async_data, "get_async_data_for_device") as mock_get:
        async_data.get_async_data_for_scan("scan_id")
        assert mock_get.call_count == 2


def test_get_async_data_for_device():
    producer = mock.MagicMock()
    async_data = AsyncDataHandler(producer)
    producer.xrange.return_value = [
        {
            "data": messages.DeviceMessage(
                signals={"data": {"value": np.zeros((10, 10))}}, metadata={}
            )
        }
    ]
    res = async_data.get_async_data_for_device("scan_id", "samx")
    assert res["data"]["value"].shape == (10, 10)
    assert len(res) == 1
    assert producer.xrange.call_count == 1
    producer.xrange.assert_called_with(
        MessageEndpoints.device_async_readback("scan_id", "samx"), min="-", max="+"
    )



import pytest
from py_scibec_openapi_client.models.beamline_with_relations import BeamlineWithRelations
from py_scibec_openapi_client.models.dataset_with_relations import DatasetWithRelations
from py_scibec_openapi_client.models.experiment_with_relations import ExperimentWithRelations
from py_scibec_openapi_client.models.scan_with_relations import ScanWithRelations

from bec_lib import bec_logger

# overwrite threads_check fixture from bec_lib,
# to have it in autouse


@pytest.fixture(autouse=True)
def threads_check(threads_check):
    yield
    bec_logger.logger.remove()


@pytest.fixture()
def active_experiment():
    return ExperimentWithRelations(
        name="dummy_experiment",
        writeAccount="p12345",
        id="dummy_experiment_id",
        beamlineId="dummy_id",
    )


@pytest.fixture()
def beamline_document():
    return BeamlineWithRelations(
        name="dummy_bl", activeExperiment="dummy_experiment", id="dummy_id"
    )


@pytest.fixture()
def dataset_document():
    return DatasetWithRelations(name="dummy_dataset", id="dummy_id")


@pytest.fixture()
def scan_document():
    return ScanWithRelations(
        name="dummy_scan",
        id="dummy_id",
        readACL=["readACL"],
        writeACL=["writeACL"],
        owner=["owner"],
    )



"""
This module provides a connector to a redis server. It is a wrapper around the
redis library providing a simple interface to send and receive messages from a
redis server.
"""

from __future__ import annotations

import collections
import inspect
import itertools
import queue
import sys
import threading
import time
import warnings
from collections.abc import MutableMapping, Sequence
from dataclasses import dataclass
from functools import wraps
from typing import TYPE_CHECKING, Optional

import louie
import redis
import redis.client
import redis.exceptions

from bec_lib.connector import ConnectorBase, MessageObject
from bec_lib.endpoints import EndpointInfo, MessageEndpoints
from bec_lib.logger import bec_logger
from bec_lib.messages import AlarmMessage, BECMessage, ClientInfoMessage, LogMessage
from bec_lib.serialization import MsgpackSerialization

if TYPE_CHECKING:
    from bec_lib.alarm_handler import Alarms


def validate_endpoint(endpoint_arg):
    def decorator(func):
        argspec = inspect.getfullargspec(func)
        argument_index = argspec.args.index(endpoint_arg)

        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                endpoint = args[argument_index]
                arg = list(args)
            except IndexError:
                endpoint = kwargs[endpoint_arg]
                arg = kwargs
            if isinstance(endpoint, str):
                warnings.warn(
                    "RedisConnector methods with a string topic are deprecated and should not be used anymore. Use RedisConnector methods with an EndpointInfo instead.",
                    DeprecationWarning,
                )
                return func(*args, **kwargs)
            elif isinstance(endpoint, EndpointInfo):
                if func.__name__ not in endpoint.message_op:
                    raise ValueError(
                        f"Endpoint {endpoint} is not compatible with {func.__name__} method"
                    )
                if isinstance(arg, list):
                    arg[argument_index] = endpoint.endpoint
                    return func(*arg, **kwargs)
                else:
                    arg[endpoint_arg] = endpoint.endpoint
                    return func(*args, **arg)
            else:
                raise TypeError(f"Endpoint {endpoint} is not EndpointInfo")

        return wrapper

    return decorator


@dataclass
class StreamSubscriptionInfo:
    id: str
    topic: str
    newest_only: bool
    from_start: bool
    cb_ref: callable
    kwargs: dict

    def __eq__(self, other):
        if not isinstance(other, StreamSubscriptionInfo):
            return False
        return (
            self.topic == other.topic
            and self.cb_ref == other.cb_ref
            and self.from_start == other.from_start
        )


@dataclass
class DirectReadingStreamSubscriptionInfo(StreamSubscriptionInfo):
    thread = None
    stop_event = None


@dataclass
class StreamMessage:
    msg: dict
    callbacks: list


class RedisConnector(ConnectorBase):
    """
    Redis connector class. This class is a wrapper around the redis library providing
    a simple interface to send and receive messages from a redis server.
    """

    def __init__(self, bootstrap: list, redis_cls=None):
        """
        Initialize the connector

        Args:
            bootstrap (list): list of strings in the form "host:port"
            redis_cls (redis.client, optional): redis client class. Defaults to None.
        """
        super().__init__(bootstrap)
        self.host, self.port = (
            bootstrap[0].split(":") if isinstance(bootstrap, list) else bootstrap.split(":")
        )

        if redis_cls:
            self._redis_conn = redis_cls(host=self.host, port=self.port)
        else:
            self._redis_conn = redis.Redis(host=self.host, port=self.port)

        # main pubsub connection
        self._pubsub_conn = self._redis_conn.pubsub()
        self._pubsub_conn.ignore_subscribe_messages = True
        # keep track of topics and callbacks
        self._topics_cb = collections.defaultdict(list)
        self._topics_cb_lock = threading.Lock()
        self._stream_topics_subscription = collections.defaultdict(list)
        self._stream_topics_subscription_lock = threading.Lock()

        self._events_listener_thread = None
        self._stream_events_listener_thread = None
        self._events_dispatcher_thread = None
        self._messages_queue = queue.Queue()
        self._stop_events_listener_thread = threading.Event()
        self._stop_stream_events_listener_thread = threading.Event()
        self.stream_keys = {}

    def shutdown(self):
        """
        Shutdown the connector
        """
        super().shutdown()
        if self._events_listener_thread:
            self._stop_events_listener_thread.set()
            self._events_listener_thread.join()
            self._events_listener_thread = None
        if self._stream_events_listener_thread:
            self._stop_stream_events_listener_thread.set()
            self._stream_events_listener_thread.join()
            self._stream_events_listener_thread = None
        if self._events_dispatcher_thread:
            self._messages_queue.put(StopIteration)
            self._events_dispatcher_thread.join()
            self._events_dispatcher_thread = None

        # this will take care of shutting down direct listening threads
        self._unregister_stream(self._stream_topics_subscription)

        # release all connections
        self._pubsub_conn.close()
        self._redis_conn.close()

    def send_client_info(
        self,
        message: str,
        rid: str = None,
        source: str = None,
        severity: int = 0,
        show_asap: bool = False,
        scope: str = None,
        metadata: dict = None,
    ):
        """
        Send a message to the client

        Args:
            msg (str): message
        """
        client_msg = ClientInfoMessage(
            message=message,
            source=source,
            severity=severity,
            show_asap=show_asap,
            scope=scope,
            RID=rid,
            metadata=metadata,
        )
        self.xadd(MessageEndpoints.client_info(), msg_dict={"data": client_msg}, max_size=100)

    def raise_alarm(self, severity: Alarms, alarm_type: str, source: str, msg: str, metadata: dict):
        """
        Raise an alarm

        Args:
            severity (Alarms): alarm severity
            alarm_type (str): alarm type
            source (str): source
            msg (str): message
            metadata (dict): metadata
        """
        alarm_msg = AlarmMessage(
            severity=severity, alarm_type=alarm_type, source=source, msg=msg, metadata=metadata
        )
        self.set_and_publish(MessageEndpoints.alarm(), alarm_msg)

    def pipeline(self) -> redis.client.Pipeline:
        """Create a new pipeline"""
        return self._redis_conn.pipeline()

    def execute_pipeline(self, pipeline) -> list:
        """
        Execute a pipeline and return the results

        Args:
            pipeline (Pipeline): redis pipeline

        Returns:
            list: list of results
        """
        if not isinstance(pipeline, redis.client.Pipeline):
            raise TypeError(f"Expected a redis Pipeline, got {type(pipeline)}")
        ret = []
        results = pipeline.execute()
        for res in results:
            try:
                ret.append(MsgpackSerialization.loads(res))
            except RuntimeError:
                ret.append(res)
        return ret

    def raw_send(self, topic: str, msg: bytes, pipe=None):
        """
        Send a message to a topic. This is the raw version of send, it does not
        check the message type. Use this method if you want to send a message
        that is not a BECMessage.

        Args:
            topic (str): topic
            msg (bytes): message
            pipe (Pipeline, optional): redis pipe. Defaults to None.
        """
        client = pipe if pipe is not None else self._redis_conn
        client.publish(topic, msg)

    @validate_endpoint("topic")
    def send(self, topic: EndpointInfo, msg: BECMessage, pipe=None) -> None:
        """
        Send a message to a topic

        Args:
            topic (str): topic
            msg (BECMessage): message
            pipe (Pipeline, optional): redis pipe. Defaults to None.
        """
        if not isinstance(msg, BECMessage):
            raise TypeError(f"Message {msg} is not a BECMessage")
        self.raw_send(topic, MsgpackSerialization.dumps(msg), pipe)

    def _start_events_dispatcher_thread(self, start_thread):
        if start_thread and self._events_dispatcher_thread is None:
            # start dispatcher thread
            started_event = threading.Event()
            self._events_dispatcher_thread = threading.Thread(
                target=self._dispatch_events, args=(started_event,)
            )
            self._events_dispatcher_thread.start()
            started_event.wait()  # synchronization of thread start

    def _convert_endpointinfo(self, endpoint, check_message_op=True):
        if isinstance(endpoint, EndpointInfo):
            return [endpoint.endpoint], endpoint.message_op.name
        if isinstance(endpoint, str):
            return [endpoint], ""
        # Support list of endpoints or dict with endpoints as keys
        if isinstance(endpoint, (Sequence, MutableMapping)):
            endpoints_str = []
            ref_message_op = None
            for e in endpoint:
                e_str, message_op = self._convert_endpointinfo(e, check_message_op=check_message_op)
                if check_message_op:
                    if ref_message_op is None:
                        ref_message_op = message_op
                    else:
                        if message_op != ref_message_op:
                            raise ValueError(
                                f"All endpoints do not have the same type: {ref_message_op}"
                            )
                endpoints_str.append(e_str)
            return list(itertools.chain(*endpoints_str)), ref_message_op or ""
        raise ValueError(f"Invalid endpoint {endpoint}")

    def _normalize_patterns(self, patterns):
        patterns, _ = self._convert_endpointinfo(patterns)
        if isinstance(patterns, str):
            return [patterns]
        elif isinstance(patterns, list):
            if not all(isinstance(p, str) for p in patterns):
                raise ValueError("register: patterns must be a string or a list of strings")
        else:
            raise ValueError("register: patterns must be a string or a list of strings")
        return patterns

    def register(
        self,
        topics: str | list[str] | EndpointInfo | list[EndpointInfo] = None,
        patterns: str | list[str] = None,
        cb: callable = None,
        start_thread: bool = True,
        from_start: bool = False,
        newest_only: bool = False,
        **kwargs,
    ):
        """
        Register a callback for a topic or a pattern

        Args:
            topics (str, list, EndpointInfo, list[EndpointInfo], optional): topic or list of topics. Defaults to None. The topic should be a valid message endpoint in BEC and can be a string or an EndpointInfo object.
            patterns (str, list, optional): pattern or list of patterns. Defaults to None. In contrast to topics, patterns may contain "*" wildcards. The evaluated patterns should be a valid pub/sub message endpoint in BEC
            cb (callable, optional): callback. Defaults to None.
            start_thread (bool, optional): start the dispatcher thread. Defaults to True.
            from_start (bool, optional): for streams only: return data from start on first reading. Defaults to False.
            newest_only (bool, optional): for streams only: return newest data only. Defaults to False.
            **kwargs: additional keyword arguments to be transmitted to the callback

        Examples:
            >>> def my_callback(msg, **kwargs):
            ...     print(msg)
            ...
            >>> connector.register("test", my_callback)
            >>> connector.register(topics="test", cb=my_callback)
            >>> connector.register(patterns="test:*", cb=my_callback)
            >>> connector.register(patterns="test:*", cb=my_callback, start_thread=False)
            >>> connector.register(patterns="test:*", cb=my_callback, start_thread=False, my_arg="test")
        """
        if cb is None:
            raise ValueError("Callback cb cannot be None")

        if topics is None and patterns is None:
            raise ValueError("topics and patterns cannot be both None")

        # make a weakref from the callable, using louie;
        # it can create safe refs for simple functions as well as methods
        cb_ref = louie.saferef.safe_ref(cb)
        item = (cb_ref, kwargs)

        if self._events_listener_thread is None:
            # create the thread that will get all messages for this connector;
            self._events_listener_thread = threading.Thread(
                target=self._get_messages_loop, args=(self._pubsub_conn,)
            )
            self._events_listener_thread.start()

        if patterns is not None:
            patterns = self._normalize_patterns(patterns)

            self._pubsub_conn.psubscribe(patterns)
            with self._topics_cb_lock:
                for pattern in patterns:
                    if item not in self._topics_cb[pattern]:
                        self._topics_cb[pattern].append(item)
        else:
            topics, message_op = self._convert_endpointinfo(topics)
            if message_op == "STREAM":
                return self._register_stream(
                    topics=topics,
                    cb=cb,
                    from_start=from_start,
                    newest_only=newest_only,
                    start_thread=start_thread,
                    **kwargs,
                )

            self._pubsub_conn.subscribe(topics)
            with self._topics_cb_lock:
                for topic in topics:
                    if item not in self._topics_cb[topic]:
                        self._topics_cb[topic].append(item)
        self._start_events_dispatcher_thread(start_thread)

    def _add_direct_stream_listener(self, topic, cb_ref, **kwargs) -> int:
        """
        Add a direct listener for a topic. This is used when newest_only is True.

        Args:
            topic (str): topic
            cb (callable): weakref to callback
            kwargs (dict): additional keyword arguments to be transmitted to the callback

        Returns:
            int: stream id
        """
        info = DirectReadingStreamSubscriptionInfo(
            id="-", topic=topic, newest_only=True, from_start=False, cb_ref=cb_ref, kwargs=kwargs
        )
        if info in self._stream_topics_subscription[topic]:
            raise RuntimeError("Already registered stream topic with the same callback")

        info.stop_event = threading.Event()
        info.thread = threading.Thread(target=self._direct_stream_listener, args=(info,))
        with self._stream_topics_subscription_lock:
            self._stream_topics_subscription[topic].append(info)
        info.thread.start()

    def _direct_stream_listener(self, info: DirectReadingStreamSubscriptionInfo):
        stop_event = info.stop_event
        cb_ref = info.cb_ref
        kwargs = info.kwargs
        topic = info.topic
        while not stop_event.is_set():
            ret = self._redis_conn.xrevrange(topic, "+", info.id, count=1)
            if not ret:
                time.sleep(0.1)
                continue
            redis_id, msg_dict = ret[0]
            timestamp, _, ind = redis_id.partition(b"-")
            info.id = f"{timestamp.decode()}-{int(ind.decode())+1}"
            stream_msg = StreamMessage(
                {key.decode(): MsgpackSerialization.loads(val) for key, val in msg_dict.items()},
                ((cb_ref, kwargs),),
            )
            self._messages_queue.put(stream_msg)

    def _get_stream_topics_id(self) -> dict:
        stream_topics_id = {}
        from_start_stream_topics_id = {}
        with self._stream_topics_subscription_lock:
            for topic, subscription_info_list in self._stream_topics_subscription.items():
                for info in subscription_info_list:
                    if isinstance(info, DirectReadingStreamSubscriptionInfo):
                        continue
                    if info.from_start:
                        from_start_stream_topics_id[topic] = info.id
                    else:
                        stream_topics_id[topic] = info.id
        return from_start_stream_topics_id, stream_topics_id

    def _handle_stream_msg_list(self, msg_list, from_start=False):
        for topic, msgs in msg_list:
            subscription_info_list = self._stream_topics_subscription[topic.decode()]
            for index, record in msgs:
                callbacks = []
                for info in subscription_info_list:
                    info.id = index.decode()
                    if from_start and not info.from_start:
                        continue
                    callbacks.append((info.cb_ref, info.kwargs))
                if callbacks:
                    msg_dict = {
                        k.decode(): MsgpackSerialization.loads(msg) for k, msg in record.items()
                    }
                    msg = StreamMessage(msg_dict, callbacks)
                    self._messages_queue.put(msg)
            for info in subscription_info_list:
                info.from_start = False

    def _get_stream_messages_loop(self) -> None:
        """
        Get stream messages loop. This method is run in a separate thread and listens
        for messages from the redis server.
        """
        error = False

        while not self._stop_stream_events_listener_thread.is_set():
            try:
                from_start_stream_topics_id, stream_topics_id = self._get_stream_topics_id()
                if not any((stream_topics_id, from_start_stream_topics_id)):
                    time.sleep(0.1)
                    continue
                msg_list = []
                from_start_msg_list = []
                # first handle the 'from_start' streams ;
                # in the case of reading from start what is expected is to call the
                # callbacks for existing items, without waiting for a new element to be added
                # to the stream
                if from_start_stream_topics_id:
                    # read the streams contents from beginning, not blocking
                    from_start_msg_list = self._redis_conn.xread(from_start_stream_topics_id)
                if stream_topics_id:
                    msg_list = self._redis_conn.xread(stream_topics_id, block=200)
            except redis.exceptions.ConnectionError:
                if not error:
                    error = True
                    bec_logger.logger.error("Failed to connect to redis. Is the server running?")
                time.sleep(1)
            # pylint: disable=broad-except
            except Exception:
                sys.excepthook(*sys.exc_info())
            else:
                error = False
                with self._stream_topics_subscription_lock:
                    self._handle_stream_msg_list(from_start_msg_list, from_start=True)
                    self._handle_stream_msg_list(msg_list)
        return True

    def _register_stream(
        self,
        topics: list[str] = None,
        cb: callable = None,
        from_start: bool = False,
        newest_only: bool = False,
        start_thread: bool = True,
        **kwargs,
    ) -> None:
        """
        Register a callback for a stream topic or pattern

        Args:
            topic (str, optional): Topic. This should be a valid message endpoint string.
            cb (callable, optional): callback. Defaults to None.
            from_start (bool, optional): read from start. Defaults to False.
            newest_only (bool, optional): read newest only. Defaults to False.
            start_thread (bool, optional): start the dispatcher thread. Defaults to True.
            **kwargs: additional keyword arguments to be transmitted to the callback

        """
        if newest_only and from_start:
            raise ValueError("newest_only and from_start cannot be both True")

        # make a weakref from the callable, using louie;
        # it can create safe refs for simple functions as well as methods
        cb_ref = louie.saferef.safe_ref(cb)

        self._start_events_dispatcher_thread(start_thread)

        if newest_only:
            # if newest_only is True, we need to provide a separate callback for each topic,
            # directly calling the callback. This is because we need to have a backpressure
            # mechanism in place, and we cannot rely on the dispatcher thread to handle it.
            for topic in topics:
                self._add_direct_stream_listener(topic, cb_ref, **kwargs)
        else:
            with self._stream_topics_subscription_lock:
                for topic in topics:
                    new_subscription = StreamSubscriptionInfo(
                        id="0-0" if from_start else "$",
                        topic=topic,
                        newest_only=newest_only,
                        from_start=from_start,
                        cb_ref=cb_ref,
                        kwargs=kwargs,
                    )
                    subscriptions = self._stream_topics_subscription[topic]
                    if new_subscription in subscriptions:
                        # raise an error if attempted to register a stream with the same callback,
                        # whereas it has already been registered as a 'direct reading' stream with
                        # newest_only=True ; it is clearly an error case that would produce weird results
                        index = subscriptions.index(new_subscription)
                        if isinstance(subscriptions[index], DirectReadingStreamSubscriptionInfo):
                            raise RuntimeError(
                                "Already registered stream topic with the same callback with 'newest_only=True'"
                            )
                    else:
                        subscriptions.append(new_subscription)

            if self._stream_events_listener_thread is None:
                # create the thread that will get all messages for this connector
                self._stream_events_listener_thread = threading.Thread(
                    target=self._get_stream_messages_loop
                )
                self._stream_events_listener_thread.start()

    def _filter_topics_cb(self, topics: list, cb: Union[callable, None]):
        unsubscribe_list = []
        with self._topics_cb_lock:
            for topic in topics:
                topics_cb = self._topics_cb[topic]
                # remove callback from list
                self._topics_cb[topic] = list(
                    filter(lambda item: cb and item[0]() is not cb, topics_cb)
                )
                if not self._topics_cb[topic]:
                    # no callbacks left, unsubscribe
                    unsubscribe_list.append(topic)
            # clean the topics that have been unsubscribed
            for topic in unsubscribe_list:
                del self._topics_cb[topic]
        return unsubscribe_list

    def unregister(self, topics=None, patterns=None, cb=None):
        if self._events_listener_thread is None:
            return

        if patterns is not None:
            patterns = self._normalize_patterns(patterns)
            # see if registered streams can be unregistered
            for pattern in patterns:
                self._unregister_stream(
                    fnmatch.filter(self._stream_topics_subscription, pattern), cb
                )
            pubsub_unsubscribe_list = self._filter_topics_cb(patterns, cb)
            if pubsub_unsubscribe_list:
                self._pubsub_conn.punsubscribe(pubsub_unsubscribe_list)
        else:
            topics, _ = self._convert_endpointinfo(topics, check_message_op=False)
            if not self._unregister_stream(topics, cb):
                unsubscribe_list = self._filter_topics_cb(topics, cb)
                if unsubscribe_list:
                    self._pubsub_conn.unsubscribe(unsubscribe_list)

    def _unregister_stream(self, topics: list[str], cb: callable = None) -> bool:
        """
        Unregister a stream listener.

        Args:
            topics (list[str]): list of stream topics

        Returns:
            bool: True if the stream listener has been removed, False otherwise
        """
        unsubscribe_list = []
        with self._stream_topics_subscription_lock:
            for topic in topics:
                subscription_infos = self._stream_topics_subscription[topic]
                # remove from list if callback corresponds
                self._stream_topics_subscription[topic] = list(
                    filter(lambda sub_info: cb and sub_info.cb_ref() is not cb, subscription_infos)
                )
                if not self._stream_topics_subscription[topic]:
                    # no callbacks left, unsubscribe
                    unsubscribe_list += subscription_infos
            # clean the topics that have been unsubscribed
            for subscription_info in unsubscribe_list:
                if isinstance(subscription_info, DirectReadingStreamSubscriptionInfo):
                    subscription_info.stop_event.set()
                    subscription_info.thread.join()
                # it is possible to register the same stream multiple times with different
                # callbacks, in this case when unregistering with cb=None (unregister all)
                # the topic can be deleted multiple times, hence try...except in code below
                try:
                    del self._stream_topics_subscription[subscription_info.topic]
                except KeyError:
                    pass

        return len(unsubscribe_list) > 0

    def _get_messages_loop(self, pubsub: redis.client.PubSub) -> None:
        """
        Get messages loop. This method is run in a separate thread and listens
        for messages from the redis server.

        Args:
            pubsub (redis.client.PubSub): pubsub object
        """
        error = False
        while not self._stop_events_listener_thread.is_set():
            try:
                msg = pubsub.get_message(timeout=1)
            except redis.exceptions.ConnectionError:
                if not error:
                    error = True
                    bec_logger.logger.error("Failed to connect to redis. Is the server running?")
                time.sleep(1)
            # pylint: disable=broad-except
            except Exception:
                sys.excepthook(*sys.exc_info())
            else:
                error = False
                if msg is not None:
                    self._messages_queue.put(msg)

    def _execute_callback(self, cb, msg, kwargs):
        try:
            cb(msg, **kwargs)
        # pylint: disable=broad-except
        except Exception:
            sys.excepthook(*sys.exc_info())

    def _handle_stream_message(self, stream_msg):
        for cb_ref, kwargs in stream_msg.callbacks:
            cb = cb_ref()
            if cb:
                self._execute_callback(cb, stream_msg.msg, kwargs)
        return True

    def _handle_message(self, msg):
        if isinstance(msg, StreamMessage):
            return self._handle_stream_message(msg)
        channel = msg["channel"].decode()
        with self._topics_cb_lock:
            if msg["pattern"] is not None:
                callbacks = self._topics_cb[msg["pattern"].decode()]
            else:
                callbacks = self._topics_cb[channel]
        msg = MessageObject(topic=channel, value=MsgpackSerialization.loads(msg["data"]))
        for cb_ref, kwargs in callbacks:
            cb = cb_ref()
            if cb:
                self._execute_callback(cb, msg, kwargs)
        return True

    def poll_messages(self, timeout=None) -> None:
        """Poll messages from the messages queue

        If timeout is None, wait for at least one message. Processes until queue is empty,
        or until timeout is reached.

        Args:

          timeout (float): timeout in seconds
        """
        start_time = time.perf_counter()
        remaining_timeout = timeout
        while True:
            try:
                # wait for a message and return it before timeout expires
                msg = self._messages_queue.get(timeout=remaining_timeout, block=True)
            except queue.Empty as exc:
                if remaining_timeout < timeout:
                    # at least one message has been processed, so we do not raise
                    # the timeout error
                    return True
                raise TimeoutError(f"{self}: timeout waiting for messages") from exc
            else:
                if msg is StopIteration:
                    return False

                self._handle_message(msg)

                if timeout is None:
                    if self._messages_queue.empty():
                        # no message to process
                        return True
                else:
                    # calculate how much time remains and retry getting a message
                    remaining_timeout = timeout - (time.perf_counter() - start_time)
                    if remaining_timeout <= 0:
                        return True

    def _dispatch_events(self, started_event):
        started_event.set()
        while self.poll_messages():
            ...

    @validate_endpoint("topic")
    def lpush(
        self, topic: EndpointInfo, msg: str, pipe=None, max_size: int = None, expire: int = None
    ) -> None:
        """Time complexity: O(1) for each element added, so O(N) to
        add N elements when the command is called with multiple arguments.
        Insert all the specified values at the head of the list stored at key.
        If key does not exist, it is created as empty list before
        performing the push operations. When key holds a value that
        is not a list, an error is returned."""
        client = pipe if pipe is not None else self.pipeline()
        if isinstance(msg, BECMessage):
            msg = MsgpackSerialization.dumps(msg)
        client.lpush(topic, msg)
        if max_size:
            client.ltrim(topic, 0, max_size)
        if expire:
            client.expire(topic, expire)
        if not pipe:
            client.execute()

    @validate_endpoint("topic")
    def lset(self, topic: EndpointInfo, index: int, msg: str, pipe=None) -> None:
        client = pipe if pipe is not None else self._redis_conn
        if isinstance(msg, BECMessage):
            msg = MsgpackSerialization.dumps(msg)
        return client.lset(topic, index, msg)

    @validate_endpoint("topic")
    def rpush(self, topic: EndpointInfo, msg: str, pipe=None) -> int:
        """O(1) for each element added, so O(N) to add N elements when the
        command is called with multiple arguments. Insert all the specified
        values at the tail of the list stored at key. If key does not exist,
        it is created as empty list before performing the push operation. When
        key holds a value that is not a list, an error is returned."""
        client = pipe if pipe is not None else self._redis_conn
        if isinstance(msg, BECMessage):
            msg = MsgpackSerialization.dumps(msg)
        return client.rpush(topic, msg)

    @validate_endpoint("topic")
    def lrange(self, topic: EndpointInfo, start: int, end: int, pipe=None):
        """O(S+N) where S is the distance of start offset from HEAD for small
        lists, from nearest end (HEAD or TAIL) for large lists; and N is the
        number of elements in the specified range. Returns the specified elements
        of the list stored at key. The offsets start and stop are zero-based indexes,
        with 0 being the first element of the list (the head of the list), 1 being
        the next element and so on."""
        client = pipe if pipe is not None else self._redis_conn
        cmd_result = client.lrange(topic, start, end)
        if pipe:
            return cmd_result

        # in case of command executed in a pipe, use 'execute_pipeline' method
        ret = []
        for msg in cmd_result:
            try:
                ret.append(MsgpackSerialization.loads(msg))
            except RuntimeError:
                ret.append(msg)
        return ret

    @validate_endpoint("topic")
    def set_and_publish(self, topic: EndpointInfo, msg, pipe=None, expire: int = None) -> None:
        """piped combination of self.publish and self.set"""
        client = pipe if pipe is not None else self.pipeline()
        if not isinstance(msg, BECMessage):
            raise TypeError(f"Message {msg} is not a BECMessage")
        msg = MsgpackSerialization.dumps(msg)
        self.set(topic, msg, pipe=client, expire=expire)
        self.raw_send(topic, msg, pipe=client)
        if not pipe:
            client.execute()

    @validate_endpoint("topic")
    def set(self, topic: EndpointInfo, msg, pipe=None, expire: int = None) -> None:
        """set redis value"""
        client = pipe if pipe is not None else self._redis_conn
        if isinstance(msg, BECMessage):
            msg = MsgpackSerialization.dumps(msg)
        client.set(topic, msg, ex=expire)

    @validate_endpoint("pattern")
    def keys(self, pattern: EndpointInfo) -> list:
        """returns all keys matching a pattern"""
        return self._redis_conn.keys(pattern)

    @validate_endpoint("topic")
    def delete(self, topic: EndpointInfo, pipe=None):
        """delete topic"""
        client = pipe if pipe is not None else self._redis_conn
        client.delete(topic)

    @validate_endpoint("topic")
    def get(self, topic: EndpointInfo, pipe=None):
        """retrieve entry, either via hgetall or get"""
        client = pipe if pipe is not None else self._redis_conn
        data = client.get(topic)
        if pipe:
            return data
        else:
            try:
                return MsgpackSerialization.loads(data)
            except RuntimeError:
                return data

    @validate_endpoint("topic")
    def xadd(
        self, topic: EndpointInfo, msg_dict: dict, max_size=None, pipe=None, expire: int = None
    ):
        """
        add to stream

        Args:
            topic (str): redis topic
            msg_dict (dict): message to add
            max_size (int, optional): max size of stream. Defaults to None.
            pipe (Pipeline, optional): redis pipe. Defaults to None.
            expire (int, optional): expire time. Defaults to None.

        Examples:
            >>> redis.xadd("test", {"test": "test"})
            >>> redis.xadd("test", {"test": "test"}, max_size=10)
        """
        if pipe:
            client = pipe
        elif expire:
            client = self.pipeline()
        else:
            client = self._redis_conn

        msg_dict = {key: MsgpackSerialization.dumps(val) for key, val in msg_dict.items()}

        if max_size:
            client.xadd(topic, msg_dict, maxlen=max_size)
        else:
            client.xadd(topic, msg_dict)
        if expire:
            client.expire(topic, expire)
        if not pipe and expire:
            client.execute()

    @validate_endpoint("topic")
    def get_last(self, topic: EndpointInfo, key=None, count=1):
        """
        Get last message from stream. Repeated calls will return
        the same message until a new message is added to the stream.

        Args:
            topic (str): redis topic
            key (str, optional): key to retrieve. Defaults to None. If None, the whole message is returned.
            count (int, optional): number of last elements to retrieve
        """
        if count <= 0:
            return None
        ret = []
        client = self._redis_conn
        try:
            res = client.xrevrange(topic, "+", "-", count=count)
            if not res:
                return None
            for _, msg_dict in reversed(res):
                ret.append(
                    {k.decode(): MsgpackSerialization.loads(msg) for k, msg in msg_dict.items()}
                    if key is None
                    else MsgpackSerialization.loads(msg_dict[key.encode()])
                )
        except TypeError:
            return None

        if count > 1:
            return ret
        else:
            return ret[0]

    @validate_endpoint("topic")
    def xread(
        self,
        topic: EndpointInfo,
        id: str = None,
        count: int = None,
        block: int = None,
        from_start=False,
    ) -> list:
        """
        read from stream

        Args:
            topic (str): redis topic
            id (str, optional): id to read from. Defaults to None.
            count (int, optional): number of messages to read. Defaults to None, which means all.
            block (int, optional): block for x milliseconds. Defaults to None.
            from_start (bool, optional): read from start. Defaults to False.

        Returns:
            [list]: list of messages

        Examples:
            >>> redis.xread("test", "0-0")
            >>> redis.xread("test", "0-0", count=1)

            # read one message at a time
            >>> key = 0
            >>> msg = redis.xread("test", key, count=1)
            >>> key = msg[0][1][0][0]
            >>> next_msg = redis.xread("test", key, count=1)
        """
        client = self._redis_conn
        if from_start:
            self.stream_keys[topic] = "0-0"
        if topic not in self.stream_keys:
            if id is None:
                try:
                    msg = client.xrevrange(topic, "+", "-", count=1)
                    if msg:
                        self.stream_keys[topic] = msg[0][0].decode()
                        out = {}
                        for key, val in msg[0][1].items():
                            out[key.decode()] = MsgpackSerialization.loads(val)
                        return [out]
                    self.stream_keys[topic] = "0-0"
                except redis.exceptions.ResponseError:
                    self.stream_keys[topic] = "0-0"
        if id is None:
            id = self.stream_keys[topic]

        msg = client.xread({topic: id}, count=count, block=block)
        return self._decode_stream_messages_xread(msg)

    def _decode_stream_messages_xread(self, msg):
        out = []
        for topic, msgs in msg:
            for index, record in msgs:
                out.append(
                    {k.decode(): MsgpackSerialization.loads(msg) for k, msg in record.items()}
                )
                self.stream_keys[topic.decode()] = index
        return out if out else None

    @validate_endpoint("topic")
    def xrange(self, topic: EndpointInfo, min: str, max: str, count: int = None):
        """
        read a range from stream

        Args:
            topic (str): redis topic
            min (str): min id. Use "-" to read from start
            max (str): max id. Use "+" to read to end
            count (int, optional): number of messages to read. Defaults to None.

        Returns:
            [list]: list of messages or None
        """
        client = self._redis_conn
        msgs = []
        for reading in client.xrange(topic, min, max, count=count):
            _, msg_dict = reading
            msgs.append(
                {k.decode(): MsgpackSerialization.loads(msg) for k, msg in msg_dict.items()}
            )
        return msgs if msgs else None

    def producer(self):
        """Return itself as a producer, to be compatible with old code"""
        warnings.warn(
            "RedisConnector.producer() is deprecated and should not be used anymore. A Connector is a producer now, just use the connector object.",
            FutureWarning,
        )
        return self

    def consumer(
        self,
        topics=None,
        patterns=None,
        group_id=None,
        event=None,
        cb=None,
        threaded=True,
        name=None,
        **kwargs,
    ):
        """Return a fake thread object to be compatible with old code

        In order to keep this fail-safe and simple it uses 'mock'...
        """
        from unittest.mock import (  # import is done here, to not pollute the file with something normally in tests
            Mock,
        )

        warnings.warn(
            "RedisConnector.consumer() is deprecated and should not be used anymore. Use RedisConnector.register() with 'topics', 'patterns', 'cb' or 'start_thread' instead. Additional keyword args are transmitted to the callback. For the caller, the main difference with RedisConnector.register() is that it does not return a new thread.",
            FutureWarning,
        )
        dummy_thread = Mock(spec=threading.Thread)
        dummy_thread.start.side_effet = lambda: self.register(
            topics, patterns, cb, threaded, **kwargs
        )
        return dummy_thread



# This file is the entry point for the BEC device server. It is responsible for
# launching the device server and handling command line arguments.
# It is called either by the bec-device-server entry point or directly from the command line.

# pylint: disable=wrong-import-position
# pylint: disable=wrong-import-order
# we need to run the startup script before we import anything else. This is
# to ensure that the epics environment variables are set correctly.
import importlib.metadata as imd

entry_points = imd.entry_points(group="bec.deployment.device_server")
for entry_point in entry_points:
    if entry_point.name == "plugin_ds_startup":
        entry_point.load()()

import argparse
import threading

from bec_lib import RedisConnector, ServiceConfig, bec_logger
from bec_server import device_server

logger = bec_logger.logger
bec_logger.level = bec_logger.LOGLEVEL.INFO


def main():
    """
    Launch the BEC device server.
    """
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--config", default="", help="path to the config file")
    clargs = parser.parse_args()
    config_path = clargs.config

    config = ServiceConfig(config_path)

    s = device_server.DeviceServer(config, RedisConnector)
    try:
        event = threading.Event()
        s.start()
        logger.success("Started DeviceServer")
        event.wait()
    except KeyboardInterrupt:
        s.shutdown()


if __name__ == "__main__":
    main()



import threading
from unittest import mock

import pytest

from bec_lib import MessageEndpoints, messages
from bec_lib.tests.utils import ConnectorMock
from bec_server.scan_server.errors import DeviceMessageError
from bec_server.scan_server.scan_stubs import ScanAbortion, ScanStubs


@pytest.fixture
def stubs():
    connector = ConnectorMock("")
    shutdown_event = threading.Event()
    yield ScanStubs(connector, shutdown_event=shutdown_event)
    shutdown_event.set()


@pytest.mark.parametrize(
    "device,parameter,metadata,reference_msg",
    [
        (
            "rtx",
            None,
            None,
            messages.DeviceInstructionMessage(
                device="rtx",
                action="kickoff",
                parameter={"configure": {}, "wait_group": "kickoff"},
                metadata={},
            ),
        ),
        (
            "rtx",
            {"num_pos": 5, "positions": [1, 2, 3, 4, 5], "exp_time": 2},
            None,
            messages.DeviceInstructionMessage(
                device="rtx",
                action="kickoff",
                parameter={
                    "configure": {"num_pos": 5, "positions": [1, 2, 3, 4, 5], "exp_time": 2},
                    "wait_group": "kickoff",
                },
                metadata={},
            ),
        ),
    ],
)
def test_kickoff(stubs, device, parameter, metadata, reference_msg):
    connector = ConnectorMock("")
    stubs = ScanStubs(connector)
    msg = list(stubs.kickoff(device=device, parameter=parameter, metadata=metadata))
    assert msg[0] == reference_msg


@pytest.mark.parametrize(
    "msg,raised_error",
    [
        (messages.DeviceRPCMessage(device="samx", return_val="", out="", success=True), None),
        (
            messages.DeviceRPCMessage(
                device="samx",
                return_val="",
                out={"error": "TypeError", "msg": "some weird error", "traceback": "traceback"},
                success=False,
            ),
            ScanAbortion,
        ),
        (
            messages.DeviceRPCMessage(device="samx", return_val="", out="", success=False),
            ScanAbortion,
        ),
    ],
)
def test_rpc_raises_scan_abortion(stubs, msg, raised_error):
    with mock.patch.object(stubs.connector, "get", return_value=msg) as prod_get:
        if raised_error is None:
            stubs._get_from_rpc("rpc-id")
        else:
            with pytest.raises(raised_error):
                stubs._get_from_rpc("rpc-id")

        prod_get.assert_called_with(MessageEndpoints.device_rpc("rpc-id"))


def test_rpc_raises_scan_abortion_on_shutdown_event(stubs):
    stubs.shutdown_event.set()
    with pytest.raises(ScanAbortion):
        stubs._get_from_rpc("rpc-id")


@pytest.mark.parametrize(
    "msg, ret_value, raised_error",
    [
        (messages.ProgressMessage(value=10, max_value=100, done=False), None, False),
        (
            messages.ProgressMessage(
                value=10, max_value=100, done=False, metadata={"RID": "wrong"}
            ),
            None,
            False,
        ),
        (
            messages.ProgressMessage(value=10, max_value=100, done=False, metadata={"RID": "rid"}),
            10,
            False,
        ),
        (
            messages.DeviceStatusMessage(device="samx", status=0, metadata={"RID": "rid"}),
            None,
            True,
        ),
    ],
)
def test_device_progress(stubs, msg, ret_value, raised_error):
    if raised_error:
        with pytest.raises(DeviceMessageError):
            with mock.patch.object(stubs.connector, "get", return_value=msg):
                assert stubs.get_device_progress(device="samx", RID="rid") == ret_value
        return
    with mock.patch.object(stubs.connector, "get", return_value=msg):
        assert stubs.get_device_progress(device="samx", RID="rid") == ret_value



"""
Scan-related utils
"""

from __future__ import annotations

import csv
import datetime
from collections import defaultdict
from typing import TYPE_CHECKING

from typeguard import typechecked

if TYPE_CHECKING:
    from bec_lib.scan_items import ScanItem
    from bec_lib.scan_report import ScanReport


@typechecked
def scan_to_csv(
    scan_report: ScanReport | ScanItem | list[ScanReport] | list[ScanItem],
    output_name: str,
    delimiter: str = ",",
    dialect: str | None = None,
    header: list | None = None,
    write_metadata: bool = True,
) -> None:
    """Convert scan data to a csv file.

    Args:
        scan_report (ScanReport | ScanItem):    ScanReport or ScanItem object(s). Can be a list of ScanReports, ScanItems
        filename (str):                         Name of the csv file.
        delimiter (str, optional):              Delimiter for the csv file. Defaults to ",".
        dialect (str, optional):                Argument for csv.Dialect. Defaults to csv.writer default, e.g. 'excel'.
                                                Other options 'excel-tab' or 'unix', still takes argument from delimiter, choose delimier='' to omit
        header (list, optional):                Create custom header for the csv file. If None, header is created automatically. Defaults to None.
        write_metadata (bool, optional):        If True, the metadata of the scan will be written to the header of csv file. Defaults to True.

    Examples:
        >>> scan_to_csv(scan_report, "./scan.csv")
    """
    if not isinstance(scan_report, list):
        scan_report = [scan_report]

    header_out = []
    body_out = []
    data_output = []

    for ii, scan_rep in enumerate(scan_report):
        if hasattr(scan_rep, "scan"):
            scan_rep = scan_rep.scan
        header_out.append(["#ScanNumber", f"{scan_rep.scan_number}"])
        header_tmp, body_tmp = _extract_scan_data(
            scan_item=scan_rep, header=header, write_metadata=write_metadata
        )
        header_out.extend(header_tmp[:-1])
        body_out.extend(body_tmp)
    header_out.append(header_tmp[-1])  # To only append the header keys once
    data_output.extend(header_out)
    data_output.extend(body_out)

    _write_csv(output_name=output_name, delimiter=delimiter, dialect=dialect, output=data_output)


def _write_csv(output_name: str, delimiter: str, output: list, dialect: str = None) -> None:
    """Write csv file.

    Args:
        output_name (str): Name of the csv file.
        delimiter (str): Delimiter for the csv file.
        dialect (str): Argument for csv.Dialect. Defaults to None. If no None, delimiter input is ignored. Some input examples 'excel', 'excel-tab' or 'unix'
        data_dict (dict): Dictionary to be written to csv.

        Examples:
            >>> _write_csv("./scan.csv", ",", ["#samx", "bpm4i"], True, scan_dict)
    """

    with open(output_name, "w", encoding="UTF-8") as file:
        writer = csv.writer(file, delimiter=delimiter, dialect=dialect)
        writer.writerows(output)


def _extract_scan_data(
    scan_item: ScanItem, header: list = None, write_metadata: bool = True
) -> tuple:
    """Extract scan data from scan report.

    Args:
        scan_item (ScanItem): ScanItem object.
        header (list, optional): Create custom header for the csv file. If None, header is created automatically. Defaults to None.
        write_metadata (bool, optional): If True, the metadata of the scan will be written to the header of csv file. Defaults to True.

    Returns:
        (tuple): Tuple of header and body of the csv file.
    """
    scan_dict = scan_to_dict(scan_item, flat=True)

    header_tmp = []
    header_tmp.append(["#scan_id", f"{scan_item.scan_id}"])
    header_tmp.append(["#ScanStatus", f"{scan_item.status}"])

    start_time = f"{datetime.datetime.fromtimestamp(scan_item.start_time).strftime('%c')}"
    header_tmp.append(["#StartTime", start_time])
    end_time = f"{datetime.datetime.fromtimestamp(scan_item.start_time).strftime('%c')}"
    header_tmp.append(["#EndTime", end_time])
    elapsed_time = (
        f"\tElapsed time: {(scan_item.end_time-scan_item.start_time):.1f} s\n"
        if scan_item.end_time and scan_item.start_time
        else ""
    )
    header_tmp.append(["#ElapsedTime", elapsed_time])

    scan_metadata = scan_item.status_message.info
    if write_metadata:
        header_tmp.append(["#ScanMetadata"])
        for key, value in scan_metadata.items():
            header_tmp.append(["".join(["#", key]), value])
    if header:
        header_keys = header
    else:
        header_keys = ["scan_number", "dataset_number"]
        # pylint: disable=expression-not-assigned
        [
            header_keys.extend([f"{value}_value", f"{time}_timestamp"])
            for value, time in zip(scan_dict["value"].keys(), scan_dict["timestamp"].keys())
        ]
    header_keys[0] = "".join(["#", header_keys[0]])
    header_tmp.append(header_keys)

    body_tmp = []
    num_entries = len(list(scan_dict["value"].values())[0])
    for ii in range(num_entries):
        sub_list = []
        sub_list.extend([scan_metadata["scan_number"], scan_metadata["dataset_number"]])
        for key in scan_dict["value"]:
            sub_list.extend([scan_dict["value"][key][ii], scan_dict["timestamp"][key][ii]])
        body_tmp.append(sub_list)

    return header_tmp, body_tmp


def scan_to_dict(scan_item: ScanItem, flat: bool = True) -> dict:
    """Convert scan data to a dictionary.

    Args:
        scan_item (ScanItem): ScanItem object.
        flat (bool, optional): If True, the dictionary will be flat. Defaults to True.

    Returns:
        (dict): Dictionary of scan data.

    Examples:
        >>> scan_to_dict(scan_report) with scan_report = scans.line_scan(...)
    """
    if flat:
        scan_dict = {"timestamp": defaultdict(lambda: []), "value": defaultdict(lambda: [])}
    else:
        scan_dict = {
            "timestamp": defaultdict(lambda: defaultdict(lambda: [])),
            "value": defaultdict(lambda: defaultdict(lambda: [])),
        }

    for dev, dev_data in scan_item.data.items():
        for signal, signal_data in dev_data.items():
            if flat:
                scan_dict["timestamp"][signal] = signal_data["timestamp"]
                scan_dict["value"][signal] = signal_data["value"]
            else:
                scan_dict["timestamp"][dev][signal] = signal_data["timestamp"]
                scan_dict["value"][dev][signal] = signal_data["value"]

    return scan_dict



import traceback

from bec_lib import bec_logger, messages

from .errors import ScanAbortion
from .scans import RequestBase, unpack_scan_args

logger = bec_logger.logger


class ScanAssembler:
    """
    ScanAssembler receives scan messages and translates the scan message into device instructions.
    """

    def __init__(self, *, parent):
        self.parent = parent
        self.device_manager = self.parent.device_manager
        self.connector = self.parent.connector
        self.scan_manager = self.parent.scan_manager

    def assemble_device_instructions(self, msg: messages.ScanQueueMessage) -> RequestBase:
        """Assemble the device instructions for a given ScanQueueMessage.
        This will be achieved by calling the specified class (must be a derived class of RequestBase)

        Args:
            msg (messages.ScanQueueMessage): scan queue message for which the instruction should be assembled

        Raises:
            ScanAbortion: Raised if the scan initialization fails.

        Returns:
            RequestBase: Scan instance of the initialized scan class
        """
        scan = msg.content.get("scan_type")
        cls_name = self.scan_manager.available_scans[scan]["class"]
        scan_cls = self.scan_manager.scan_dict[cls_name]

        logger.info(f"Preparing instructions of request of type {scan} / {scan_cls.__name__}")
        try:
            args = unpack_scan_args(msg.content.get("parameter", {}).get("args", []))
            kwargs = msg.content.get("parameter", {}).get("kwargs", {})
            scan_instance = scan_cls(
                *args,
                device_manager=self.device_manager,
                parameter=msg.content.get("parameter"),
                metadata=msg.metadata,
                **kwargs,
            )
            return scan_instance
        except Exception as exc:
            content = traceback.format_exc()
            logger.error(
                f"Failed to initialize the scan class of type {scan_cls.__name__}. {content}"
            )
            raise ScanAbortion from exc



import copy
import os
from unittest import mock

import pytest

import bec_lib
from bec_lib import messages
from bec_server.device_server.devices.config_update_handler import ConfigUpdateHandler
from bec_server.device_server.devices.devicemanager import DeviceConfigError, DeviceManagerDS

dir_path = os.path.dirname(bec_lib.__file__)


@pytest.mark.parametrize("device_manager_class", [DeviceManagerDS])
def test_request_response(session_from_test_config, device_manager):
    def get_config_from_mock():
        device_manager._session = copy.deepcopy(session_from_test_config)
        device_manager._load_session()

    def mocked_failed_connection(obj):
        if obj.name == "samx":
            raise ConnectionError

    config_reply = messages.RequestResponseMessage(accepted=True, message="")
    with mock.patch.object(device_manager, "connect_device", wraps=mocked_failed_connection):
        with mock.patch.object(device_manager, "_get_config", get_config_from_mock):
            with mock.patch.object(
                device_manager.config_helper, "wait_for_config_reply", return_value=config_reply
            ):
                with mock.patch.object(device_manager.config_helper, "wait_for_service_response"):
                    device_manager.initialize("")
                    with mock.patch.object(
                        device_manager.config_update_handler, "send_config_request_reply"
                    ) as request_reply:
                        device_manager.config_update_handler.parse_config_request(
                            msg=messages.DeviceConfigMessage(
                                action="update", config={"something": "something"}
                            )
                        )
                        request_reply.assert_called_once()


@pytest.mark.parametrize("device_manager_class", [DeviceManagerDS])
def test_config_handler_update_config(dm_with_devices):
    device_manager = dm_with_devices
    handler = ConfigUpdateHandler(device_manager)

    # bpm4i doesn't have a controller, so it should be destroyed
    msg = messages.DeviceConfigMessage(action="update", config={"bpm4i": {"enabled": False}})
    handler._update_config(msg)
    assert device_manager.devices.bpm4i.enabled is False
    assert device_manager.devices.bpm4i.initialized is False
    assert device_manager.devices.bpm4i.obj._destroyed is True

    msg = messages.DeviceConfigMessage(action="update", config={"bpm4i": {"enabled": True}})
    handler._update_config(msg)
    assert device_manager.devices.bpm4i.enabled is True
    assert device_manager.devices.bpm4i.initialized is True
    assert device_manager.devices.bpm4i.obj._destroyed is False


@pytest.mark.parametrize("device_manager_class", [DeviceManagerDS])
def test_config_handler_update_config_raises(dm_with_devices):
    device_manager = dm_with_devices
    handler = ConfigUpdateHandler(device_manager)

    msg = messages.DeviceConfigMessage(
        action="update", config={"samx": {"deviceConfig": {"doesntexist": True}}}
    )
    old_config = device_manager.devices.samx._config["deviceConfig"].copy()
    with pytest.raises(DeviceConfigError):
        handler._update_config(msg)
    assert device_manager.devices.samx._config["deviceConfig"] == old_config


@pytest.mark.parametrize("device_manager_class", [DeviceManagerDS])
def test_reload_action(dm_with_devices):
    device_manager = dm_with_devices
    handler = ConfigUpdateHandler(device_manager)
    dm = handler.device_manager
    with mock.patch.object(dm.devices.samx.obj, "destroy") as obj_destroy:
        with mock.patch.object(dm, "_get_config") as get_config:
            handler._reload_config()
            obj_destroy.assert_called_once()
            get_config.assert_called_once()


@pytest.mark.parametrize("device_manager_class", [DeviceManagerDS])
def test_parse_config_request_update(dm_with_devices):
    handler = ConfigUpdateHandler(dm_with_devices)
    msg = messages.DeviceConfigMessage(
        action="update", config={"samx": {"deviceConfig": {"doesntexist": True}}}
    )
    with mock.patch.object(handler, "_update_config") as update_config:
        handler.parse_config_request(msg)
        update_config.assert_called_once_with(msg)


@pytest.mark.parametrize("device_manager_class", [DeviceManagerDS])
def test_parse_config_request_reload(device_manager):
    handler = ConfigUpdateHandler(device_manager)
    dm = handler.device_manager
    dm.failed_devices = ["samx"]
    msg = messages.DeviceConfigMessage(action="reload", config={})
    with mock.patch.object(handler, "_reload_config") as reload_config:
        handler.parse_config_request(msg)
        reload_config.assert_called_once()
        assert msg.metadata["failed_devices"] == ["samx"]



import threading
import time as ttime

import numpy as np
from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from ophyd import Device, OphydObject, PositionerBase


class DummyControllerDevice(Device):
    USER_ACCESS = ["controller"]


class DummyController:
    USER_ACCESS = [
        "some_var",
        "some_var_property",
        "controller_show_all",
        "_func_with_args",
        "_func_with_args_and_kwargs",
        "_func_with_kwargs",
        "_func_without_args_kwargs",
    ]
    some_var = 10
    another_var = 20

    def __init__(self) -> None:
        self._some_var_property = None
        self.connected = False

    @property
    def some_var_property(self):
        return self._some_var_property

    def on(self):
        self.connected = True

    def off(self):
        self.connected = False

    def _func_with_args(self, *args):
        return args

    def _func_with_args_and_kwargs(self, *args, **kwargs):
        return args, kwargs

    def _func_with_kwargs(self, **kwargs):
        return kwargs

    def _func_without_args_kwargs(self):
        return None

    def controller_show_all(self):
        """dummy controller show all

        Raises:
            in: _description_
            LimitError: _description_

        Returns:
            _type_: _description_
        """
        print(self.some_var)


class SynController(OphydObject):
    def on(self):
        pass

    def off(self):
        pass


class SynFlyerLamNI(Device, PositionerBase):
    def __init__(
        self,
        *,
        name,
        readback_func=None,
        value=0,
        delay=0,
        speed=1,
        update_frequency=2,
        precision=3,
        parent=None,
        labels=None,
        kind=None,
        device_manager=None,
        **kwargs,
    ):
        if readback_func is None:

            def readback_func(x):
                return x

        self.sim_state = {}
        self._readback_func = readback_func
        self.delay = delay
        self.precision = precision
        self.tolerance = kwargs.pop("tolerance", 0.5)
        self.device_manager = device_manager

        # initialize values
        self.sim_state["readback"] = readback_func(value)
        self.sim_state["readback_ts"] = ttime.time()

        super().__init__(name=name, parent=parent, labels=labels, kind=kind, **kwargs)
        self.controller = SynController(name="SynController")

    def kickoff(self, metadata, num_pos, positions, exp_time: float = 0):
        positions = np.asarray(positions)

        def produce_data(device, metadata):
            buffer_time = 0.2
            elapsed_time = 0
            bundle = messages.BundleMessage()
            for ii in range(num_pos):
                bundle.append(
                    messages.DeviceMessage(
                        signals={
                            "flyer_samx": {"value": positions[ii, 0], "timestamp": 0},
                            "flyer_samy": {"value": positions[ii, 1], "timestamp": 0},
                        },
                        metadata={"point_id": ii, **metadata},
                    )
                )
                ttime.sleep(exp_time)
                elapsed_time += exp_time
                if elapsed_time > buffer_time:
                    elapsed_time = 0
                    device.device_manager.connector.set_and_publish(
                        MessageEndpoints.device_read(device.name), bundle
                    )
                    bundle = messages.BundleMessage()
                    device.device_manager.connector.set(
                        MessageEndpoints.device_status(device.name),
                        messages.DeviceStatusMessage(
                            device=device.name, status=1, metadata={"point_id": ii, **metadata}
                        ),
                    )
            device.device_manager.connector.send(MessageEndpoints.device_read(device.name), bundle)
            device.device_manager.connector.set(
                MessageEndpoints.device_status(device.name),
                messages.DeviceStatusMessage(
                    device=device.name, status=0, metadata={"point_id": num_pos, **metadata}
                ),
            )
            print("done")

        flyer = threading.Thread(target=produce_data, args=(self, metadata))
        flyer.start()



# pylint: skip-file
import os
from unittest import mock

import numpy as np
import pytest

import bec_lib
from bec_lib import MessageEndpoints, ServiceConfig, messages
from bec_lib.logger import bec_logger
from bec_lib.redis_connector import MessageObject
from bec_lib.tests.utils import ConnectorMock
from bec_server.file_writer import FileWriterManager
from bec_server.file_writer.file_writer import FileWriter
from bec_server.file_writer.file_writer_manager import ScanStorage

# pylint: disable=missing-function-docstring
# pylint: disable=protected-access

dir_path = os.path.dirname(bec_lib.__file__)


@pytest.fixture
def file_writer_manager_mock():
    connector_cls = ConnectorMock
    config = ServiceConfig(
        redis={"host": "dummy", "port": 6379},
        config={
            "file_writer": {"plugin": "default_NeXus_format", "base_path": "./"},
            "log_writer": {"base_path": "./"},
        },
    )
    with (
        mock.patch.object(FileWriterManager, "_start_device_manager", return_value=None),
        mock.patch.object(FileWriterManager, "wait_for_service"),
    ):
        file_writer_manager_mock = FileWriterManager(config=config, connector_cls=connector_cls)
        try:
            yield file_writer_manager_mock
        finally:
            file_writer_manager_mock.shutdown()
            bec_logger.logger.remove()
            bec_logger._reset_singleton()


def test_scan_segment_callback(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    msg = messages.ScanMessage(
        point_id=1, scan_id="scan_id", data={"data": "data"}, metadata={"scan_number": 1}
    )
    msg_bundle = messages.BundleMessage()
    msg_bundle.append(msg)
    msg_raw = MessageObject(value=msg_bundle, topic="scan_segment")

    file_manager._scan_segment_callback(msg_raw, parent=file_manager)
    assert file_manager.scan_storage["scan_id"].scan_segments[1] == {"data": "data"}


def test_scan_status_callback(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    msg = messages.ScanStatusMessage(
        scan_id="scan_id",
        status="closed",
        info={
            "scan_number": 1,
            "DIID": "DIID",
            "stream": "stream",
            "scan_type": "step",
            "num_points": 1,
            "enforce_sync": True,
        },
    )
    msg_raw = MessageObject(value=msg, topic="scan_status")

    file_manager._scan_status_callback(msg_raw, parent=file_manager)
    assert file_manager.scan_storage["scan_id"].scan_finished is True


class MockWriter(FileWriter):
    def __init__(self, file_writer_manager):
        super().__init__(file_writer_manager)
        self.write_called = False

    def write(self, file_path: str, data):
        self.write_called = True


def test_write_file(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    file_manager.scan_storage["scan_id"] = ScanStorage(10, "scan_id")
    with mock.patch.object(
        file_manager.writer_mixin, "compile_full_filename"
    ) as mock_create_file_path:
        mock_create_file_path.return_value = "path"
        # replace NexusFileWriter with MockWriter
        file_manager.file_writer = MockWriter(file_manager)
        file_manager.write_file("scan_id")
        assert file_manager.file_writer.write_called is True


def test_write_file_invalid_scan_id(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    file_manager.scan_storage["scan_id"] = ScanStorage(10, "scan_id")
    with mock.patch.object(
        file_manager.writer_mixin, "compile_full_filename"
    ) as mock_create_file_path:
        file_manager.write_file("scan_id1")
        mock_create_file_path.assert_not_called()


def test_write_file_invalid_scan_number(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    file_manager.scan_storage["scan_id"] = ScanStorage(10, "scan_id")
    file_manager.scan_storage["scan_id"].scan_number = None
    with mock.patch.object(
        file_manager.writer_mixin, "compile_full_filename"
    ) as mock_create_file_path:
        file_manager.write_file("scan_id")
        mock_create_file_path.assert_not_called()


def test_write_file_raises_alarm_on_error(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    file_manager.scan_storage["scan_id"] = ScanStorage(10, "scan_id")
    with mock.patch.object(
        file_manager.writer_mixin, "compile_full_filename"
    ) as mock_compile_filename:
        with mock.patch.object(file_manager, "connector") as mock_connector:
            mock_compile_filename.return_value = "path"
            # replace NexusFileWriter with MockWriter
            file_manager.file_writer = MockWriter(file_manager)
            file_manager.file_writer.write = mock.Mock(side_effect=Exception("error"))
            file_manager.write_file("scan_id")
            mock_connector.raise_alarm.assert_called_once()


def test_update_baseline_reading(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    file_manager.scan_storage["scan_id"] = ScanStorage(10, "scan_id")
    with mock.patch.object(file_manager, "connector") as mock_connector:
        mock_connector.get.return_value = messages.ScanBaselineMessage(
            scan_id="scan_id", data={"data": "data"}
        )
        file_manager.update_baseline_reading("scan_id")
        assert file_manager.scan_storage["scan_id"].baseline == {"data": "data"}
        mock_connector.get.assert_called_once_with(MessageEndpoints.public_scan_baseline("scan_id"))


def test_scan_storage_append():
    storage = ScanStorage(10, "scan_id")
    storage.append(1, {"data": "data"})
    assert storage.scan_segments[1] == {"data": "data"}
    assert storage.scan_finished is False


def test_scan_storage_ready_to_write():
    storage = ScanStorage(10, "scan_id")
    storage.num_points = 1
    storage.scan_finished = True
    storage.append(1, {"data": "data"})
    assert storage.ready_to_write() is True


def test_update_file_references(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    with mock.patch.object(file_manager, "connector") as mock_connector:
        file_manager.update_file_references("scan_id")
        mock_connector.keys.assert_not_called()


def test_update_file_references_gets_keys(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    file_manager.scan_storage["scan_id"] = ScanStorage(10, "scan_id")
    with mock.patch.object(file_manager, "connector") as mock_connector:
        file_manager.update_file_references("scan_id")
        mock_connector.keys.assert_called_once_with(MessageEndpoints.public_file("scan_id", "*"))


def test_update_async_data(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    file_manager.scan_storage["scan_id"] = ScanStorage(10, "scan_id")
    with mock.patch.object(file_manager, "connector") as mock_connector:
        with mock.patch.object(file_manager, "_process_async_data") as mock_process:
            key = MessageEndpoints.device_async_readback("scan_id", "dev1").endpoint
            mock_connector.keys.return_value = [key.encode()]
            data = [(b"0-0", b'{"data": "data"}')]
            mock_connector.xrange.return_value = data
            file_manager.update_async_data("scan_id")
            mock_connector.xrange.assert_called_once_with(key, min="-", max="+")
            mock_process.assert_called_once_with(data, "scan_id", "dev1")


def test_process_async_data_single_entry(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    file_manager.scan_storage["scan_id"] = ScanStorage(10, "scan_id")
    data = [{"data": messages.DeviceMessage(signals={"data": {"value": np.zeros((10, 10))}})}]
    file_manager._process_async_data(data, "scan_id", "dev1")
    assert np.isclose(
        file_manager.scan_storage["scan_id"].async_data["dev1"]["data"]["value"], np.zeros((10, 10))
    ).all()


def test_process_async_data_extend(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    file_manager.scan_storage["scan_id"] = ScanStorage(10, "scan_id")
    data = [
        {
            "data": messages.DeviceMessage(
                signals={"data": {"value": np.zeros((10, 10))}}, metadata={"async_update": "extend"}
            )
        }
        for ii in range(10)
    ]
    file_manager._process_async_data(data, "scan_id", "dev1")
    assert file_manager.scan_storage["scan_id"].async_data["dev1"]["data"]["value"].shape == (
        100,
        10,
    )


def test_process_async_data_append(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    file_manager.scan_storage["scan_id"] = ScanStorage(10, "scan_id")
    data = [
        {
            "data": messages.DeviceMessage(
                signals={"data": {"value": np.zeros((10, 10))}}, metadata={"async_update": "append"}
            )
        }
        for ii in range(10)
    ]
    file_manager._process_async_data(data, "scan_id", "dev1")
    assert len(file_manager.scan_storage["scan_id"].async_data["dev1"]["data"]) == 10


def test_process_async_data_replace(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    file_manager.scan_storage["scan_id"] = ScanStorage(10, "scan_id")
    data = [
        {
            "data": messages.DeviceMessage(
                signals={"data": {"value": np.zeros((10, 10))}},
                metadata={"async_update": "replace"},
            )
        }
        for ii in range(10)
    ]
    file_manager._process_async_data(data, "scan_id", "dev1")
    assert file_manager.scan_storage["scan_id"].async_data["dev1"]["data"]["value"].shape == (
        10,
        10,
    )


def test_update_scan_storage_with_status_ignores_none(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    file_manager.update_scan_storage_with_status(
        messages.ScanStatusMessage(scan_id=None, status="closed", info={})
    )
    assert file_manager.scan_storage == {}


def test_ready_to_write(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    file_manager.scan_storage["scan_id"] = ScanStorage(10, "scan_id")
    file_manager.scan_storage["scan_id"].scan_finished = True
    file_manager.scan_storage["scan_id"].num_points = 1
    file_manager.scan_storage["scan_id"].scan_segments = {"0": {"data": np.zeros((10, 10))}}
    assert file_manager.scan_storage["scan_id"].ready_to_write() is True
    file_manager.scan_storage["scan_id1"] = ScanStorage(101, "scan_id1")
    file_manager.scan_storage["scan_id1"].scan_finished = True
    file_manager.scan_storage["scan_id1"].num_points = 2
    file_manager.scan_storage["scan_id1"].scan_segments = {"0": {"data": np.zeros((10, 10))}}
    assert file_manager.scan_storage["scan_id1"].ready_to_write() is False


def test_ready_to_write_forced(file_writer_manager_mock):
    file_manager = file_writer_manager_mock
    file_manager.scan_storage["scan_id"] = ScanStorage(10, "scan_id")
    file_manager.scan_storage["scan_id"].scan_finished = False
    file_manager.scan_storage["scan_id"].forced_finish = True
    assert file_manager.scan_storage["scan_id"].ready_to_write() is True



"""
This module contains the classes for storing scan data from scan_segments. 
"""

import collections
from typing import Any

from _collections_abc import dict_items, dict_keys, dict_values

from bec_lib import messages


class SignalData:
    """
    SignalData is a container for storing signal data.
    """

    def __init__(self):
        self.metadata = {}
        self.scan_id = None
        self.num_points = None
        self.data = {}

    def __getitem__(self, key: Any) -> Any:
        if isinstance(key, int):
            return self.data.get(key)
        if key in ["val", "value"]:
            return self.val
        if key == "timestamp":
            return self.timestamps
        return self.get(key)

    @property
    def val(self):
        """return a list of values of the signal data"""
        return [self.data.get(index, {}).get("value") for index in sorted(self.data.keys())]

    @property
    def timestamps(self):
        """return a list of timestamps of the signal data"""
        return [self.data.get(index, {}).get("timestamp") for index in sorted(self.data.keys())]

    def get(self, index: Any, default=None) -> dict:
        """
        Get the signal data at the given index.

        Args:
            index(int): the index of the scan point

        Returns:
            dict: the signal data at the given index
        """
        if isinstance(index, int):
            return self.data.get(index, default)
        if index in ["val", "value"]:
            return self.val
        if index == "timestamp":
            return self.timestamps
        return self.get(index, default)

    def set(self, index: Any, device_data: dict) -> None:
        """
        Set the signal data at the given index to the given device data.

        Args:
            index(int): the index of the scan point
            device_data(dict): the device data to store

        """

        self.data[index] = device_data

    def __eq__(self, __value: object) -> bool:
        return self.data == __value

    def items(self) -> dict_items:
        return self.data.items()

    def keys(self) -> dict_keys:
        return self.data.keys()

    def values(self) -> dict_values:
        return self.data.values()

    def val_to_dict(self) -> dict:
        return {k: v.get("value") for k, v in self.data.items()}

    def __str__(self) -> str:
        return f"{self.data}"

    def __len__(self) -> int:
        return len(self.data)


class DeviceData(dict):
    """
    DeviceData is a container for storing device data.
    """

    def __init__(self):
        self.__signals = collections.defaultdict(SignalData)
        super().__init__()

    def __getitem__(self, key: Any) -> Any:
        if key in self.__signals:
            return self.__signals[key]
        if isinstance(key, int):
            return {signal: val.get(key) for signal, val in self.__signals.items()}
        return self.get(key)

    def __setattr__(self, attr: Any, value: Any) -> None:
        self.__setitem__(attr, value)

    def __setitem__(self, __key: Any, __value: Any) -> None:
        super().__setitem__(__key, __value)
        self.__dict__.update({__key: __value})

    def __delitem__(self, __key: Any) -> None:
        super().__delitem__(__key)
        del self.__dict__[__key]

    def get(self, index: Any, default=None) -> Any:
        if index in self.__signals:
            return self.__signals[index]
        if isinstance(index, int):
            return {signal: val.get(index) for signal, val in self.__signals.items()}
        return self.__signals.get(index, default)

    def set(self, index: Any, signals: dict) -> None:
        for signal, signal_data in signals.items():
            self.__signals[signal].set(index, signal_data)
            self.__setattr__(signal, self.__signals[signal])

    def __str__(self) -> str:
        return f"{dict(self.__signals)}"

    def __eq__(self, ref_data: object) -> bool:
        return {name: self.__signals[name].data for name in self.__signals} == ref_data

    def keys(self) -> dict_keys:
        return self.__signals.keys()

    def items(self) -> dict_items:
        return self.__signals.items()

    def values(self) -> dict_values:
        return self.__signals.values()


class ScanData(dict):
    """
    ScanData is a container for storing scan data.
    """

    def __init__(self, *args, **kwargs):
        self.devices = collections.defaultdict(DeviceData)
        self.messages = collections.defaultdict()
        super().__init__(*args, **kwargs)

    def get(self, index: Any, default=None) -> Any:
        if index in self.devices:
            return self.devices.get(index, default)
        if isinstance(index, int) and index in self.messages:
            return self.messages.get(index)
        return self.devices.get(index, default)

    def __getitem__(self, key: Any) -> Any:
        if key in self.devices:
            return self.devices[key]
        return self.get(key)

    def __contains__(self, key: Any) -> bool:
        if key in self.devices:
            return True
        return False

    def __eq__(self, __value: object) -> bool:
        return super().__eq__(__value)

    def __setattr__(self, attr: Any, value: Any) -> None:
        self.__setitem__(attr, value)

    def __setitem__(self, __key: Any, __value: Any) -> None:
        super().__setitem__(__key, __value)
        self.__dict__.update({__key: __value})

    def __delitem__(self, __key: Any) -> None:
        super().__delitem__(__key)
        del self.__dict__[__key]

    def set(self, index: Any, message: messages.BECMessage) -> None:
        """
        Set the scan data at the given index to the values in the given content.

        Args:
            index(int): the index of the scan point
            content(dict): the content to store

        """
        if not isinstance(index, int):
            raise TypeError("ScanData can only store data with integer indices.")

        self.messages[index] = message
        for dev, dev_data in message.content["data"].items():
            self.devices[dev].set(index, dev_data)
            self.__setattr__(dev, self.devices[dev])

    def keys(self) -> dict_keys:
        return self.devices.keys()

    def items(self) -> dict_items:
        return self.devices.items()

    def values(self) -> dict_values:
        return self.devices.values()

    def __str__(self) -> str:
        return f"{dict(self.devices)}"

    def __len__(self) -> int:
        return len(self.messages)



import io

import yaml


def include_constructor(loader, node):
    """
    Include another yaml file.
    """
    filename = loader.construct_scalar(node)
    with open(filename, "r", encoding="utf-8") as file_in:
        return {"__include__": {"data": yaml_load(file_in), "filename": filename}}


def bec_loader():
    """
    Returns a yaml loader that can include other yaml files using the !include tag.
    """
    loader = yaml.Loader
    loader.add_constructor("!include", include_constructor)
    return loader


def yaml_load(stream: io.TextIOWrapper | str) -> dict:
    """
    Load a yaml file with the ability to include other yaml files.

    Args:
        stream (io.TextIOWrapper | str): The yaml file to load. Can be a file object or a string pointing to a file.

    Returns:
        dict: The yaml file as a dictionary.
    """

    if isinstance(stream, str):
        with open(stream, "r", encoding="utf-8") as file_in:
            return _parse_data_stream(file_in)
    else:
        return _parse_data_stream(stream)


def _parse_data_stream(stream: io.TextIOWrapper) -> dict:
    out = yaml.load(stream, Loader=bec_loader())
    included = []
    for k, v in out.items():
        if "__include__" in v:
            included.append((k, v))
        elif isinstance(v, list):
            for item in v:
                if "__include__" in item:
                    included.append((k, item))
    for k, v in included:
        if k in out:
            out.pop(k)
        for key, value in v["__include__"]["data"].items():
            if key in out:
                print(
                    f"Warning: Multiple definitions for key {key}. Using the one from {v['__include__']['filename']}."
                )
            out[key] = value
    return out



"""
This module contains the AsyncDataHandler class which is used to receive and store async device data from the BEC.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from bec_lib.endpoints import MessageEndpoints

if TYPE_CHECKING:
    from bec_lib import messages
    from bec_lib.connector import ConnectorBase


class AsyncDataHandler:
    def __init__(self, connector: ConnectorBase):
        self.connector = connector

    def get_async_data_for_scan(self, scan_id: str) -> dict[list]:
        """
        Get the async data for a given scan.

        Args:
            scan_id(str): the scan id to get the async data for

        Returns:
            dict[list]: the async data for the scan sorted by device name
        """
        async_device_keys = self.connector.keys(
            MessageEndpoints.device_async_readback(scan_id, "*")
        )
        async_data = {}
        for device_key in async_device_keys:
            key = device_key.decode()
            device_name = key.split(MessageEndpoints.device_async_readback(scan_id, "").endpoint)[
                -1
            ].split(":")[0]
            data = self.get_async_data_for_device(scan_id, device_name)
            if not data:
                continue
            async_data[device_name] = data
        return async_data

    def get_async_data_for_device(self, scan_id: str, device_name: str) -> list:
        """
        Get the async data for a given device in a scan.

        Args:
            scan_id(str): the scan id to get the async data for
            device_name(str): the device name to get the async data for

        Returns:
            list: the async data for the device
        """
        key = MessageEndpoints.device_async_readback(scan_id, device_name)
        msgs = self.connector.xrange(key, min="-", max="+")
        if not msgs:
            return []
        return self.process_async_data(msgs)

    @staticmethod
    def process_async_data(msgs: list[messages.DeviceMessage]) -> dict | list[dict]:
        """
        Process the async data.

        Args:
            device_name(str): the name of the device
            msgs(list[messages.DeviceMessage]): the async data to process

        Returns:
            list: the processed async data
        """
        concat_type = None
        data = []
        async_data = {}
        for msg in msgs:
            msg = msg["data"]
            if not concat_type:
                concat_type = msg.metadata.get("async_update", "append")
            data.append(msg.content["signals"])
        if len(data) == 1:
            async_data = data[0]
            return async_data
        if concat_type == "extend":
            # concatenate the dictionaries
            for signal in data[0].keys():
                async_data[signal] = {}
                for key in data[0][signal].keys():
                    if hasattr(data[0][signal][key], "__iter__"):
                        async_data[signal][key] = np.concatenate([d[signal][key] for d in data])
                    else:
                        async_data[signal][key] = [d[signal][key] for d in data]
            return async_data
        if concat_type == "append":
            # concatenate the lists
            for key in data[0].keys():
                async_data[key] = []
                for d in data:
                    async_data[key].append(d[key])
            return async_data
        if concat_type == "replace":
            # replace the dictionaries
            async_data = data[-1]
            return async_data
        raise ValueError(f"Unknown async update type: {concat_type}")



"""
This module provides a connector to the electronic logbook SciLog.
"""

from __future__ import annotations

import sys
import warnings
from typing import TYPE_CHECKING

from requests.exceptions import HTTPError

from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger

logger = bec_logger.logger

try:
    import scilog
except ImportError:
    logger.info("Unable to import `scilog` optional dependency")

if TYPE_CHECKING:
    from bec_lib.redis_connector import RedisConnector


class LogbookConnector:
    def __init__(self, connector: RedisConnector) -> None:
        self.connector = connector
        self.connected = False
        self._scilog_module = None
        self._connect()
        self.logbook = None

    def _connect(self):
        if "scilog" not in sys.modules:
            return

        msg = self.connector.get(MessageEndpoints.logbook())
        if not msg:
            return
        scilog_creds = msg.credentials

        account_msg = self.connector.get(MessageEndpoints.account())
        if not account_msg:
            return
        account = account_msg.value
        account = account.replace("e", "p")

        self._scilog_module = scilog

        self.log = self._scilog_module.SciLog(
            scilog_creds["url"], options={"token": scilog_creds["token"]}
        )
        # FIXME the python sdk should not use the ownergroup
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            try:
                logbooks = self.log.get_logbooks(readACL={"inq": [account]})
            except HTTPError:
                self.connector.set(MessageEndpoints.logbook(), b"")
                return
        if len(logbooks) > 1:
            logger.warning("Found two logbooks. Taking the first one.")
        self.log.select_logbook(logbooks[0])

        # set aliases
        # pylint: disable=no-member, invalid-name
        self.LogbookMessage = self._scilog_module.LogbookMessage
        self.send_logbook_message = self.log.send_logbook_message
        self.send_message = self.log.send_message

        self.connected = True


# if __name__ == "__main__":
#     import datetime

#     logbook = LogbookConnector()
#     msg = LogbookMessage(logbook)
#     msg.add_text(
#         f"<p><mark class='pen-red'><strong>Beamline checks failed at {str(datetime.datetime.now())}.</strong></mark></p>"
#     ).add_tag("BEC")
#     logbook.send_logbook_message(msg)



from __future__ import annotations

from bec_lib import Alarms, BECService, BECStatus
from bec_lib import DeviceManagerBase as DeviceManager
from bec_lib import MessageEndpoints, ServiceConfig, bec_logger, messages
from bec_lib.connector import ConnectorBase

from .scan_assembler import ScanAssembler
from .scan_guard import ScanGuard
from .scan_manager import ScanManager
from .scan_queue import QueueManager

logger = bec_logger.logger


class ScanServer(BECService):
    device_manager = None
    queue_manager = None
    scan_guard = None
    scan_server = None
    scan_assembler = None
    scan_manager = None

    def __init__(self, config: ServiceConfig, connector_cls: ConnectorBase):
        super().__init__(config, connector_cls, unique_service=True)
        self._start_scan_manager()
        self._start_queue_manager()
        self._start_device_manager()
        self._start_scan_guard()
        self._start_scan_assembler()
        # self._start_scan_server()
        self._start_alarm_handler()
        self._reset_scan_number()
        self.status = BECStatus.RUNNING

    def _start_device_manager(self):
        self.wait_for_service("DeviceServer")
        self.device_manager = DeviceManager(self)
        self.device_manager.initialize([self.bootstrap_server])

    def _start_scan_manager(self):
        self.scan_manager = ScanManager(parent=self)

    def _start_queue_manager(self):
        self.queue_manager = QueueManager(parent=self)

    def _start_scan_assembler(self):
        self.scan_assembler = ScanAssembler(parent=self)

    def _start_scan_guard(self):
        self.scan_guard = ScanGuard(parent=self)

    def _start_alarm_handler(self):
        self.connector.register(MessageEndpoints.alarm(), cb=self._alarm_callback, parent=self)

    def _reset_scan_number(self):
        if self.connector.get(MessageEndpoints.scan_number()) is None:
            self.scan_number = 1
        if self.connector.get(MessageEndpoints.dataset_number()) is None:
            self.dataset_number = 1

    @staticmethod
    def _alarm_callback(msg, parent: ScanServer, **_kwargs):
        msg = msg.value
        queue = msg.metadata.get("queue", "primary")
        if Alarms(msg.content["severity"]) == Alarms.MAJOR:
            logger.info(f"Received alarm: {msg}")
            parent.queue_manager.set_abort(queue=queue)

    @property
    def scan_number(self) -> int:
        """get the current scan number"""
        msg = self.connector.get(MessageEndpoints.scan_number())
        if msg is None:
            logger.warning("Failed to retrieve scan number from redis.")
            return -1
        if not isinstance(msg, messages.VariableMessage):
            # This is a temporary fix for the scan number being stored as a string
            # in the redis database. This will be removed in the future.
            num = int(msg)
            self.scan_number = num
            return num
        return int(msg.value)

    @scan_number.setter
    def scan_number(self, val: int):
        """set the current scan number"""
        msg = messages.VariableMessage(value=val)
        self.connector.set(MessageEndpoints.scan_number(), msg)

    @property
    def dataset_number(self) -> int:
        """get the current dataset number"""
        msg = self.connector.get(MessageEndpoints.dataset_number())
        if msg is None:
            logger.warning("Failed to retrieve dataset number from redis.")
            return -1
        if not isinstance(msg, messages.VariableMessage):
            # This is a temporary fix for the dataset number being stored as a string
            # in the redis database. This will be removed in the future.
            num = int(msg)
            self.dataset_number = num
            return num
        return int(msg.value)

    @dataset_number.setter
    def dataset_number(self, val: int):
        """set the current dataset number"""
        msg = messages.VariableMessage(value=val)
        self.connector.set(MessageEndpoints.dataset_number(), msg)

    def shutdown(self) -> None:
        """shutdown the scan server"""
        self.device_manager.shutdown()
        self.queue_manager.shutdown()



import time

import numpy as np
import pytest
import yaml

from bec_lib import DeviceConfigError, bec_logger
from bec_lib.alarm_handler import AlarmBase

logger = bec_logger.logger


@pytest.mark.timeout(100)
def test_grid_scan_lib(bec_client_lib):
    bec = bec_client_lib
    scans = bec.scans
    bec.metadata.update({"unit_test": "test_grid_scan_bec_client_lib"})
    dev = bec.device_manager.devices
    scans.umv(dev.samx, 0, dev.samy, 0, relative=False)
    status = scans.grid_scan(dev.samx, -5, 5, 10, dev.samy, -5, 5, 10, exp_time=0.01, relative=True)
    status.wait()
    assert len(status.scan.data) == 100
    assert status.scan.num_points == 100


@pytest.mark.timeout(100)
def test_mv_scan_lib(bec_client_lib):
    bec = bec_client_lib
    scans = bec.scans
    bec.metadata.update({"unit_test": "test_mv_scan_bec_client_lib"})
    dev = bec.device_manager.devices
    scans.mv(dev.samx, 10, dev.samy, 20, relative=False).wait()
    current_pos_samx = dev.samx.read()["samx"]["value"]
    current_pos_samy = dev.samy.read()["samy"]["value"]
    assert np.isclose(
        current_pos_samx, 10, atol=dev.samx._config["deviceConfig"].get("tolerance", 0.05)
    )
    assert np.isclose(
        current_pos_samy, 20, atol=dev.samy._config["deviceConfig"].get("tolerance", 0.05)
    )


@pytest.mark.timeout(100)
def test_mv_raises_limit_error(bec_client_lib):
    bec = bec_client_lib
    scans = bec.scans
    bec.metadata.update({"unit_test": "test_mv_raises_limit_error"})
    dev = bec.device_manager.devices
    dev.samx.limits = [-50, 50]
    with pytest.raises(AlarmBase) as exc:
        scans.mv(dev.samx, 1000, relative=False).wait()


@pytest.mark.timeout(100)
def test_async_callback_data_matches_scan_data_lib(bec_client_lib):
    bec = bec_client_lib
    scans = bec.scans  # not needed but to silence pylint...
    bec.metadata.update({"unit_test": "test_async_callback_data_matches_scan_data"})
    dev = bec.device_manager.devices
    reference_container = {"data": [], "metadata": {}}

    def dummy_callback(data, metadata):
        logger.info(f"callback metadata: {metadata}")
        reference_container["metadata"] = metadata
        reference_container["data"].append(data)

    s = scans.line_scan(dev.samx, 0, 1, steps=10, relative=False, async_callback=dummy_callback)
    s.wait()
    while len(reference_container["data"]) < 10:
        time.sleep(0.1)
    assert len(s.scan.data) == 10
    assert len(reference_container["data"]) == 10

    for ii, msg in enumerate(s.scan.data.messages.values()):
        assert msg.content == reference_container["data"][ii]


@pytest.mark.timeout(100)
def test_config_updates(bec_client_lib):
    bec = bec_client_lib
    bec.metadata.update({"unit_test": "test_config_updates"})
    dev = bec.device_manager.devices
    dev.samx.limits = [-80, 80]
    assert dev.samx.limits == [-80, 80]
    dev.samx.limits = [-50, 50]
    assert dev.samx.limits == [-50, 50]

    dev.samx.velocity.set(10).wait()
    assert dev.samx.velocity.read()["samx_velocity"]["value"] == 10
    assert dev.samx.velocity.read(cached=False)["samx_velocity"]["value"] == 10
    assert dev.samx.read_configuration()["samx_velocity"]["value"] == 10
    assert dev.samx.read_configuration(cached=False)["samx_velocity"]["value"] == 10

    dev.samx.velocity.put(5)
    assert dev.samx.velocity.get() == 5

    dev.samx.velocity.set(10).wait()
    assert dev.samx.velocity.get() == 10

    dev.samx.setpoint.put(5)
    assert dev.samx.setpoint.get() == 5

    dev.samx.setpoint.set(10).wait()
    assert dev.samx.setpoint.get() == 10
    assert dev.samx.dummy_controller.some_var == 10
    dev.samx.dummy_controller.some_var = 20
    assert dev.samx.dummy_controller.some_var == 20
    dev.samx.dummy_controller.some_var = 10
    val = dev.samx.readback.get()
    assert np.isclose(val, dev.samx.position, atol=0.05)


@pytest.mark.timeout(100)
def test_dap_fit(bec_client_lib):
    bec = bec_client_lib
    bec.metadata.update({"unit_test": "test_dap_fit"})
    dev = bec.device_manager.devices
    scans = bec.scans

    dev.bpm4i.sim.sim_select_model("GaussianModel")
    params = dev.bpm4i.sim.sim_params
    params.update(
        {"noise": "uniform", "noise_multiplier": 10, "center": 5, "sigma": 1, "amplitude": 200}
    )
    dev.bpm4i.sim.sim_params = params
    time.sleep(1)

    res = scans.line_scan(dev.samx, 0, 8, steps=50, relative=False)
    res.wait()

    fit = bec.dap.GaussianModel.fit(res.scan, "samx", "samx", "bpm4i", "bpm4i")

    assert np.isclose(fit.center, 5, atol=0.5)


@pytest.mark.timeout(100)
@pytest.mark.parametrize(
    "config, raises_error, deletes_config, disabled_device",
    [
        (
            {
                "hexapod": {
                    "deviceClass": "ophyd_devices.SynDeviceOPAAS",
                    "deviceConfig": {},
                    "deviceTags": ["user motors"],
                    "readoutPriority": "baseline",
                    "enabled": True,
                    "readOnly": False,
                },
                "eyefoc": {
                    "deviceClass": "ophyd_devices.SimPositioner",
                    "deviceConfig": {
                        "delay": 1,
                        "limits": [-50, 50],
                        "speed": 100,
                        "tolerance": 0.01,
                        "update_frequency": 400,
                    },
                    "deviceTags": ["user motors"],
                    "enabled": True,
                    "readOnly": False,
                },
            },
            True,
            False,
            [],
        ),
        (
            {
                "hexapod": {
                    "deviceClass": "ophyd_devices.SynDeviceOPAAS",
                    "deviceConfig": {},
                    "deviceTags": ["user motors"],
                    "readoutPriority": "baseline",
                    "enabled": True,
                    "readOnly": False,
                },
                "eyefoc": {
                    "deviceClass": "ophyd_devices.SimPositioner",
                    "deviceConfig": {
                        "delay": 1,
                        "limits": [-50, 50],
                        "speed": 100,
                        "tolerance": 0.01,
                        "update_frequency": 400,
                    },
                    "readoutPriority": "baseline",
                    "deviceTags": ["user motors"],
                    "enabled": True,
                    "readOnly": False,
                },
            },
            False,
            False,
            [],
        ),
        (
            {
                "hexapod": {
                    "deviceClass": "ophyd_devices.SynDeviceOPAAS",
                    "deviceConfig": {},
                    "deviceTags": ["user motors"],
                    "readoutPriority": "baseline",
                    "enabled": True,
                    "readOnly": False,
                },
                "eyefoc": {
                    "deviceClass": "ophyd_devices.utils.bec_utils.DeviceClassConnectionError",
                    "deviceConfig": {},
                    "readoutPriority": "baseline",
                    "deviceTags": ["user motors"],
                    "enabled": True,
                    "readOnly": False,
                },
            },
            True,
            False,
            ["eyefoc"],
        ),
        (
            {
                "hexapod": {
                    "deviceClass": "SynDeviceOPAAS",
                    "deviceConfig": {},
                    "deviceTags": ["user motors"],
                    "readoutPriority": "baseline",
                    "enabled": True,
                    "readOnly": False,
                },
                "eyefoc": {
                    "deviceClass": "ophyd_devices.utils.bec_utils.DeviceClassInitError",
                    "deviceConfig": {},
                    "readoutPriority": "baseline",
                    "deviceTags": ["user motors"],
                    "enabled": True,
                    "readOnly": False,
                },
            },
            True,
            True,
            [],
        ),
        (
            {
                "hexapod": {
                    "deviceClass": "SynDeviceOPAAS",
                    "deviceConfig": {},
                    "deviceTags": ["user motors"],
                    "readoutPriority": "baseline",
                    "enabled": True,
                    "readOnly": False,
                },
                "eyefoc": {
                    "deviceClass": "ophyd_devices.WrongDeviceClass",
                    "deviceConfig": {},
                    "readoutPriority": "baseline",
                    "deviceTags": ["user motors"],
                    "enabled": True,
                    "readOnly": False,
                },
            },
            True,
            True,
            [],
        ),
    ],
    ids=[
        "invalid_config_missing_readoutPriority",
        "valid_config_no_error",
        "invalid_device_class_connection_error",
        "invalid_device_class_init",
        "invalid_device_class",
    ],
)
def test_config_reload(
    bec_test_config_file_path, bec_client_lib, config, raises_error, deletes_config, disabled_device
):
    bec = bec_client_lib
    bec.metadata.update({"unit_test": "test_config_reload"})
    runtime_config_file_path = bec_test_config_file_path.parent / "e2e_runtime_config_test.yaml"
    try:
        # write new config to disk
        with open(runtime_config_file_path, "w") as f:
            f.write(yaml.dump(config))
        num_devices = len(bec.device_manager.devices)
        if raises_error:
            with pytest.raises(DeviceConfigError):
                bec.config.update_session_with_file(runtime_config_file_path)
            if deletes_config:
                assert len(bec.device_manager.devices) == 0
            elif disabled_device:
                assert len(bec.device_manager.devices) == 2
            else:
                assert len(bec.device_manager.devices) == num_devices
        else:
            bec.config.update_session_with_file(runtime_config_file_path)
            assert len(bec.device_manager.devices) == 2
        for dev in disabled_device:
            assert bec.device_manager.devices[dev].enabled is False
    finally:
        bec.config.update_session_with_file(bec_test_config_file_path)


def test_computed_signal(bec_client_lib):
    bec = bec_client_lib
    bec.metadata.update({"unit_test": "test_computed_signal"})
    dev = bec.device_manager.devices
    scans = bec.scans

    res = scans.line_scan(dev.samx, -0.1, 0.1, steps=10, relative=False, exp_time=0)
    res.wait()
    assert "pseudo_signal1" in res.scan.baseline

    def compute_signal1(*args, **kwargs):
        return 5

    dev.pseudo_signal1.set_compute_method(compute_signal1)
    dev.pseudo_signal1.set_input_signals()

    assert dev.pseudo_signal1.read(cached=False)["pseudo_signal1"]["value"] == 5



from unittest import mock

import pytest

from bec_lib import MessageEndpoints, messages
from bec_lib.connector import MessageObject

# pylint: disable=missing-function-docstring
# pylint: disable=protected-access


@pytest.fixture()
def dummy_signal_data():
    return {
        "samx": {"value": 0.51, "timestamp": 1234.56},
        "samx_setpoint": {"value": 0.51, "timestamp": 1234.56},
    }


@pytest.fixture()
def dummy_device_data_message(dummy_signal_data):
    return messages.DeviceMessage(
        signals=dummy_signal_data, metadata={"scan_id": "scan_id", "readout_priority": "monitored"}
    )


def test_device_read_callback(scan_bundler_mock, dummy_signal_data):
    dev_msg = messages.DeviceMessage(
        signals=dummy_signal_data, metadata={"scan_id": "laksjd", "readout_priority": "monitored"}
    )
    msg = MessageObject(MessageEndpoints.device_read("samx").endpoint, dev_msg)

    with mock.patch.object(scan_bundler_mock, "_add_device_to_storage") as add_dev:
        scan_bundler_mock._device_read_callback(msg)
        add_dev.assert_called_once_with([dev_msg], "samx")


@pytest.mark.parametrize(
    "scan_id,storageID,scan_msg",
    [
        ("adlk-jalskdj", None, None),
        (
            "adlk-jalskdjs",
            "adlk-jalskdjs",
            messages.ScanStatusMessage(
                scan_id="adlk-jalskdjs",
                status="open",
                info={
                    "scan_motors": ["samx"],
                    "readout_priority": {"monitored": ["samx"], "baseline": [], "on_request": []},
                    "queue_id": "my-queue-ID",
                    "scan_number": 5,
                    "scan_type": "step",
                },
            ),
        ),
        (
            "adlk-jalskdjs",
            "",
            messages.ScanStatusMessage(
                scan_id="adlk-jalskdjs",
                status="open",
                info={
                    "scan_motors": ["samx"],
                    "readout_priority": {"monitored": ["samx"], "baseline": [], "on_request": []},
                    "queue_id": "my-queue-ID",
                    "scan_number": 5,
                    "scan_type": "step",
                },
            ),
        ),
    ],
)
def test_wait_for_scan_id(scan_bundler_mock, scan_id, storageID, scan_msg):
    sb = scan_bundler_mock
    sb.storage_initialized.add(storageID)
    with mock.patch.object(sb.connector, "get", return_value=scan_msg) as get_scan_msgs:
        if not storageID and not scan_msg:
            with pytest.raises(TimeoutError):
                sb._wait_for_scan_id(scan_id, 1)
            return
        sb._wait_for_scan_id(scan_id)


def test_add_device_to_storage_returns_without_scan_id(scan_bundler_mock, dummy_signal_data):
    msg = messages.DeviceMessage(
        signals=dummy_signal_data, metadata={"readout_priority": "monitored"}
    )
    sb = scan_bundler_mock
    sb._add_device_to_storage([msg], "samx", timeout_time=1)
    assert "samx" not in sb.device_storage


def test_add_device_to_storage_returns_without_signal(scan_bundler_mock):
    msg = messages.DeviceMessage(
        signals={}, metadata={"scan_id": "scan_id", "readout_priority": "monitored"}
    )
    sb = scan_bundler_mock
    sb._add_device_to_storage([msg], "samx", timeout_time=1)
    assert "samx" not in sb.device_storage


def test_add_device_to_storage_returns_on_timeout(scan_bundler_mock, dummy_signal_data):
    msg = messages.DeviceMessage(
        signals=dummy_signal_data, metadata={"scan_id": "scan_id", "readout_priority": "monitored"}
    )
    sb = scan_bundler_mock
    sb._add_device_to_storage([msg], "samx", timeout_time=1)
    assert "samx" not in sb.device_storage


@pytest.mark.parametrize("scan_status", ["aborted", "closed"])
def test_add_device_to_storage_returns_without_scan_info(
    scan_bundler_mock, scan_status, dummy_signal_data
):
    msg = messages.DeviceMessage(
        signals=dummy_signal_data, metadata={"scan_id": "scan_id", "readout_priority": "monitored"}
    )
    sb = scan_bundler_mock
    sb.sync_storage["scan_id"] = {"info": {}}
    sb.sync_storage["scan_id"]["status"] = scan_status
    sb._add_device_to_storage([msg], "samx", timeout_time=1)
    assert "samx" not in sb.device_storage


@pytest.mark.parametrize(
    "msg,scan_type",
    [
        ("dummy_device_data_message", "step"),
        ("dummy_device_data_message", "fly"),
        ("dummy_device_data_message", "wrong"),
    ],
)
def test_add_device_to_storage_primary(scan_bundler_mock, msg, scan_type, request):
    msg = request.getfixturevalue(msg)
    sb = scan_bundler_mock
    sb.sync_storage["scan_id"] = {"info": {"scan_type": scan_type, "monitor_sync": "bec"}}
    sb.sync_storage["scan_id"]["status"] = "open"
    sb.monitored_devices["scan_id"] = {"devices": [sb.device_manager.devices.samx]}
    sb.storage_initialized.add("scan_id")
    if scan_type == "step":
        with mock.patch.object(sb, "_step_scan_update") as step_update:
            sb._add_device_to_storage([msg], "samx", timeout_time=1)
            step_update.assert_called_once_with(
                "scan_id", "samx", msg.content["signals"], msg.metadata
            )
        return
    if scan_type == "fly":
        with mock.patch.object(sb, "_fly_scan_update") as fly_update:
            sb._add_device_to_storage([msg], "samx", timeout_time=1)
            fly_update.assert_called_once_with(
                "scan_id", "samx", msg.content["signals"], msg.metadata
            )
        return
    with pytest.raises(RuntimeError):
        sb._add_device_to_storage([msg], "samx", timeout_time=1)


@pytest.mark.parametrize(
    "msg,scan_type",
    [
        (
            messages.DeviceMessage(
                signals={
                    "samx": {"value": 0.51},
                    "setpoint": {"value": 0.5},
                    "motor_is_moving": {"value": 0},
                },
                metadata={"scan_id": "scan_id"},
            ),
            "fly",
        ),
        (
            messages.DeviceMessage(
                signals={
                    "flyer": {"value": 0.51},
                    "flyer_setpoint": {"value": 0.5},
                    "flyer_motor_is_moving": {"value": 0},
                },
                metadata={"scan_id": "scan_id"},
            ),
            "fly",
        ),
    ],
)
def test_add_device_to_storage_primary_flyer(scan_bundler_mock, msg, scan_type):
    sb = scan_bundler_mock
    sb.sync_storage["scan_id"] = {"info": {"scan_type": scan_type, "monitor_sync": "flyer"}}
    sb.sync_storage["scan_id"]["status"] = "open"
    sb.storage_initialized.add("scan_id")
    sb.monitored_devices["scan_id"] = {"devices": [sb.device_manager.devices.samx], "point_id": {}}
    sb.readout_priority["scan_id"] = {
        "monitored": [],
        "baseline": [],
        "on_request": [],
        "triggering_master": "flyer",
    }
    with mock.patch.object(sb, "_fly_scan_update") as fly_update:
        sb._add_device_to_storage([msg], "samx", timeout_time=1)
        fly_update.assert_called_once_with("scan_id", "samx", msg.content["signals"], msg.metadata)
    return


@pytest.mark.parametrize(
    "msg,scan_type",
    [
        (
            messages.DeviceMessage(
                signals={
                    "samx": {"value": 0.51},
                    "setpoint": {"value": 0.5},
                    "motor_is_moving": {"value": 0},
                },
                metadata={"scan_id": "scan_id", "readout_priority": "baseline"},
            ),
            "step",
        )
    ],
)
def test_add_device_to_storage_baseline(scan_bundler_mock, msg, scan_type):
    sb = scan_bundler_mock
    sb.sync_storage["scan_id"] = {"info": {"scan_type": scan_type, "monitor_sync": "bec"}}
    sb.sync_storage["scan_id"]["status"] = "open"
    sb.monitored_devices["scan_id"] = {"devices": []}
    sb.storage_initialized.add("scan_id")
    with mock.patch.object(sb, "_baseline_update") as step_update:
        sb._add_device_to_storage([msg], "samx", timeout_time=1)
        step_update.assert_called_once_with("scan_id", "samx", msg.content["signals"])


@pytest.mark.parametrize(
    "queue_msg",
    [
        messages.ScanQueueStatusMessage(
            queue={
                "primary": {
                    "info": [
                        {
                            "queue_id": "7c15c9a2-71d4-4f2a-91a7-c4a63088fa38",
                            "scan_id": ["bfa582aa-f9cd-4258-ab5d-3e5d54d3dde5"],
                            "is_scan": [True],
                            "request_blocks": [
                                {
                                    "msg": b"\x84\xa8msg_type\xa4scan\xa7content\x83\xa9scan_type\xabfermat_scan\xa9parameter\x82\xa4args\x82\xa4samx\x92\xfe\x02\xa4samy\x92\xfe\x02\xa6kwargs\x83\xa4step\xcb?\xf8\x00\x00\x00\x00\x00\x00\xa8exp_time\xcb?\x94z\xe1G\xae\x14{\xa8relative\xc3\xa5queue\xa7primary\xa8metadata\x81\xa3RID\xd9$cd8fc68f-fe65-4031-9a37-e0e7ba9df542\xa7version\xcb?\xf0\x00\x00\x00\x00\x00\x00",
                                    "RID": "cd8fc68f-fe65-4031-9a37-e0e7ba9df542",
                                    "scan_motors": ["samx", "samy"],
                                    "readout_priority": {
                                        "monitored": ["samx", "samy"],
                                        "baseline": [],
                                        "on_request": [],
                                    },
                                    "is_scan": True,
                                    "scan_number": 25,
                                    "scan_id": "bfa582aa-f9cd-4258-ab5d-3e5d54d3dde5",
                                    "metadata": {"RID": "cd8fc68f-fe65-4031-9a37-e0e7ba9df542"},
                                    "content": {
                                        "scan_type": "fermat_scan",
                                        "parameter": {
                                            "args": {"samx": [-2, 2], "samy": [-2, 2]},
                                            "kwargs": {
                                                "step": 1.5,
                                                "exp_time": 0.02,
                                                "relative": True,
                                            },
                                        },
                                        "queue": "primary",
                                    },
                                }
                            ],
                            "scan_number": [25],
                            "status": "PENDING",
                            "active_request_block": None,
                        }
                    ],
                    "status": "RUNNING",
                }
            }
        )
    ],
)
def test_scan_queue_callback(scan_bundler_mock, queue_msg):
    sb = scan_bundler_mock
    msg = MessageObject("", queue_msg)
    sb._scan_queue_callback(msg)
    assert sb.current_queue == queue_msg.content["queue"]["primary"].get("info")


@pytest.mark.parametrize(
    "scan_msg",
    [
        messages.ScanStatusMessage(
            scan_id="6ff7a89a-79e5-43ad-828b-c1e1aeed5803",
            status="closed",
            info={
                "readout_priority": "monitored",
                "DIID": 4,
                "RID": "a53538b4-79f3-4132-91b5-d044e438f460",
                "scan_id": "3ea07f69-b0ee-44fa-8451-b85824a37397",
                "queue_id": "84e5bc19-e2fc-4b03-b706-004420322813",
                "primary": ["samx", "samy"],
                "num_points": 143,
            },
        )
    ],
)
def test_scan_status_callback(scan_bundler_mock, scan_msg):
    sb = scan_bundler_mock
    msg = MessageObject("", scan_msg)

    with mock.patch.object(sb, "handle_scan_status_message") as handle_scan_status_message_mock:
        sb._scan_status_callback(msg)
        handle_scan_status_message_mock.assert_called_once_with(scan_msg)


@pytest.mark.parametrize(
    "scan_msg, sync_storage",
    [
        [
            messages.ScanStatusMessage(
                scan_id="6ff7a89a-79e5-43ad-828b-c1e1aeed5803",
                status="closed",
                info={
                    "readout_priority": "monitored",
                    "DIID": 4,
                    "RID": "a53538b4-79f3-4132-91b5-d044e438f460",
                    "scan_id": "3ea07f69-b0ee-44fa-8451-b85824a37397",
                    "queue_id": "84e5bc19-e2fc-4b03-b706-004420322813",
                    "primary": ["samx", "samy"],
                    "num_points": 143,
                },
            ),
            [],
        ],
        [
            messages.ScanStatusMessage(
                scan_id="6ff7a89a-79e5-43ad-828b-c1e1aeed5803",
                status="open",
                info={
                    "readout_priority": "monitored",
                    "DIID": 4,
                    "RID": "a53538b4-79f3-4132-91b5-d044e438f460",
                    "scan_id": "3ea07f69-b0ee-44fa-8451-b85824a37397",
                    "queue_id": "84e5bc19-e2fc-4b03-b706-004420322813",
                    "primary": ["samx", "samy"],
                    "num_points": 143,
                },
            ),
            ["6ff7a89a-79e5-43ad-828b-c1e1aeed5803"],
        ],
    ],
)
def test_handle_scan_status_message(scan_bundler_mock, scan_msg, sync_storage):
    sb = scan_bundler_mock
    scan_id = scan_msg.content["scan_id"]
    sb.sync_storage = sync_storage

    with mock.patch.object(sb, "cleanup_storage") as cleanup_storage_mock:
        with mock.patch.object(sb, "_initialize_scan_container") as init_mock:
            with mock.patch.object(sb, "_scan_status_modification") as status_mock:
                sb.handle_scan_status_message(scan_msg)
                if scan_id not in sb.sync_storage:
                    init_mock.assert_called_once_with(scan_msg)
                    assert scan_id in sb.scan_id_history
                else:
                    init_mock.assert_not_called()

                if scan_msg.content.get("status") != "open":
                    status_mock.assert_called_once_with(scan_msg)
                else:
                    status_mock.assert_not_called()


def test_status_modification(scan_bundler_mock):
    scan_id = "test_scan_id"
    scan_bundler_mock.sync_storage[scan_id] = {"status": "open"}
    msg = messages.ScanStatusMessage(
        scan_id=scan_id,
        status="closed",
        info={
            "primary": ["samx"],
            "queue_id": "my-queue-ID",
            "scan_number": 5,
            "scan_type": "step",
        },
    )
    scan_bundler_mock._scan_status_modification(msg)
    assert scan_bundler_mock.sync_storage[scan_id]["status"] == "closed"

    scan_id = "scan_id_not_available"
    msg = messages.ScanStatusMessage(
        scan_id=scan_id,
        status="closed",
        info={
            "primary": ["samx"],
            "queue_id": "my-queue-ID",
            "scan_number": 5,
            "scan_type": "step",
        },
    )
    scan_bundler_mock._scan_status_modification(msg)
    assert scan_bundler_mock.sync_storage[scan_id]["info"] == {}


@pytest.mark.parametrize(
    "scan_msg",
    [
        messages.ScanStatusMessage(
            scan_id="6ff7a89a-79e5-43ad-828b-c1e1aeed5803",
            status="closed",
            info={
                "DIID": 4,
                "RID": "a53538b4-79f3-4132-91b5-d044e438f460",
                "scan_id": "3ea07f69-b0ee-44fa-8451-b85824a37397",
                "queue_id": "84e5bc19-e2fc-4b03-b706-004420322813",
                "scan_number": 5,
                "scan_motors": ["samx", "samy"],
                "readout_priority": {
                    "monitored": ["samx", "samy"],
                    "baseline": [],
                    "on_request": [],
                },
                "num_points": 143,
            },
        ),
        messages.ScanStatusMessage(
            scan_id="6ff7a89a-79e5-43ad-828b-c1e1aeed5803",
            status="open",
            info={
                "DIID": 4,
                "RID": "a53538b4-79f3-4132-91b5-d044e438f460",
                "scan_id": "3ea07f69-b0ee-44fa-8451-b85824a37397",
                "queue_id": "84e5bc19-e2fc-4b03-b706-004420322813",
                "scan_number": 5,
                "scan_motors": ["samx", "samy", "eyex", "bpm3a"],
                "readout_priority": {
                    "monitored": ["samx", "samy", "eyex", "bpm3a"],
                    "baseline": [],
                    "on_request": [],
                },
                "num_points": 143,
            },
        ),
    ],
)
def test_initialize_scan_container(scan_bundler_mock, scan_msg):
    sb = scan_bundler_mock
    scan_id = scan_msg.content["scan_id"]
    scan_info = scan_msg.content["info"]
    scan_motors = list(set(sb.device_manager.devices[m] for m in scan_info["scan_motors"]))
    readout_priority = scan_info["readout_priority"]
    bl_devs = sb.device_manager.devices.baseline_devices(readout_priority=readout_priority)

    with mock.patch.object(sb, "run_emitter") as emitter_mock:
        sb._initialize_scan_container(
            scan_msg
        )  # The sb.device_manager.devices[m] will crash if m is not a motor in devices

        if scan_msg.content.get("status") != "open":
            return
        assert sb.scan_motors[scan_id] == scan_motors
        assert sb.sync_storage[scan_id] == {"info": scan_info, "status": "open", "sent": set()}
        assert sb.monitored_devices[scan_id] == {
            "devices": sb.device_manager.devices.monitored_devices(
                readout_priority=readout_priority
            ),
            "point_id": {},
        }
        assert "eyex" not in [dev.name for dev in bl_devs]
        assert sb.baseline_devices[scan_id] == {
            "devices": bl_devs,
            "done": {dev.name: False for dev in bl_devs},
        }

        assert scan_id in sb.storage_initialized
        emitter_mock.assert_called_once_with("on_init", scan_id)


@pytest.mark.parametrize(
    "scan_msg, point_id, primary",
    [
        [
            messages.DeviceMessage(
                signals={"samx": {"value": 0.51, "timestamp": 1234.56}},
                metadata={
                    "scan_id": "adlk-jalskdja",
                    "readout_priority": "monitored",
                    "point_id": 23,
                },
            ),
            23,
            True,
        ],
        [
            messages.DeviceMessage(
                signals={"samx": {"value": 0.51, "timestamp": 1234.56}},
                metadata={
                    "scan_id": "adlk-jalskdjb",
                    "readout_priority": "monitored",
                    "point_id": 23,
                },
            ),
            23,
            False,
        ],
        [
            messages.DeviceMessage(
                signals={"samx": {"value": 0.51, "timestamp": 1234.56}},
                metadata={"scan_id": "adlk-jalskdjc", "readout_priority": "monitored"},
            ),
            23,
            False,
        ],
    ],
)
def test_step_scan_update(scan_bundler_mock, scan_msg, point_id, primary):
    sb = scan_bundler_mock

    metadata = scan_msg.metadata
    scan_id = metadata.get("scan_id")
    device = "samx"
    signal = scan_msg.content.get("signals")
    sb.sync_storage[scan_id] = {"info": {}, "status": "open", "sent": set()}
    scan_motors = list(set(sb.device_manager.devices[m] for m in ["samx", "samy"]))

    monitored_devices = sb.monitored_devices[scan_id] = {
        "devices": sb.device_manager.devices.monitored_devices(scan_motors),
        "point_id": {},
    }

    dev = {device: signal}
    if primary:
        monitored_devices["point_id"][point_id] = {
            dev.name: True for dev in monitored_devices["devices"]
        }

    with mock.patch.object(sb, "_update_monitor_signals") as update_mock:
        with mock.patch.object(sb, "_send_scan_point") as send_mock:
            sb._step_scan_update(scan_id, device, signal, metadata)

            if "point_id" not in metadata:
                assert sb.sync_storage[scan_id] == {"info": {}, "status": "open", "sent": set()}
                return

            assert sb.sync_storage[scan_id][point_id] == {
                **sb.sync_storage[scan_id].get(point_id, {}),
                **dev,
            }

            assert monitored_devices["point_id"][point_id][device] == True

            if primary:
                update_mock.assert_called_once()
                send_mock.assert_called_once()

            else:
                pd_test = {dev.name: False for dev in monitored_devices["devices"]}
                pd_test["samx"] = True
                assert monitored_devices["point_id"][point_id] == pd_test


@pytest.mark.parametrize(
    "scan_id,storage,remove",
    [
        ("lkasjd", {"status": "open"}, False),
        ("alskjd", {"status": "closed"}, True),
        ("poiflkj", {"status": "aborted"}, True),
    ],
)
def test_cleanup_storage(scan_bundler_mock, scan_id, storage, remove):
    sb = scan_bundler_mock
    sb.sync_storage[scan_id] = storage
    sb.storage_initialized.add(scan_id)
    with mock.patch.object(sb, "run_emitter") as emitter:
        sb.cleanup_storage()
        if remove:
            emitter.assert_called_once_with("on_cleanup", scan_id)
            assert scan_id not in sb.storage_initialized
        else:
            emitter.assert_not_called()
            assert scan_id in sb.storage_initialized


@pytest.mark.parametrize("scan_id,point_id,sent", [("lkasjd", 1, True), ("alskjd", 2, False)])
def test_send_scan_point(scan_bundler_mock, scan_id, point_id, sent):
    sb = scan_bundler_mock
    sb.sync_storage[scan_id] = {"sent": set([1])}
    sb.sync_storage[scan_id][point_id] = {}
    with mock.patch.object(sb, "run_emitter") as emitter:
        with mock.patch("bec_server.scan_bundler.scan_bundler.logger") as logger:
            sb._send_scan_point(scan_id, point_id)
            emitter.assert_called_once_with("on_scan_point_emit", scan_id, point_id)
            if sent:
                logger.debug.assert_called_once()


def test_run_emitter(scan_bundler_mock):
    sb = scan_bundler_mock
    with mock.patch("bec_server.scan_bundler.scan_bundler.logger") as logger:
        sb.run_emitter("on_init", "jlaksjd", "jlkasjd")
        logger.error.assert_called()

    with mock.patch.object(sb._emitter[0], "on_init") as init:
        sb.run_emitter("on_init", "jlaksjd")
        init.assert_called_once_with("jlaksjd")


@pytest.mark.parametrize(
    "scan_id,device,signal,metadata",
    [
        ("scan_id-lkjd", "bpm4r", {"value": 5}, {"point_id": 2}),
        ("scan_id-lkjd", "bpm4r", {"value": 5}, {}),
    ],
)
def test_fly_scan_update(scan_bundler_mock, scan_id, device, signal, metadata):
    sb = scan_bundler_mock
    sb.sync_storage[scan_id] = {}
    with mock.patch.object(sb, "_update_monitor_signals") as update_signals:
        with mock.patch.object(sb, "_send_scan_point") as send_point:
            sb.sync_storage[scan_id]["info"] = {"monitor_sync": "flyer"}
            sb._fly_scan_update(scan_id, device, signal, metadata)
            point_id = metadata.get("point_id")
            if point_id:
                update_signals.assert_called_once_with(scan_id, point_id)
                send_point.assert_called_once_with(scan_id, point_id)


@pytest.mark.parametrize("scan_id,device,signal", [("scan_id-lkjd", "bpm4r", {"value": 5})])
def test_baseline_update(scan_bundler_mock, scan_id, device, signal):
    sb = scan_bundler_mock
    sb.baseline_devices[scan_id] = {"done": {device: False}}
    sb.sync_storage[scan_id] = {}
    sb.scan_motors[scan_id] = []
    sb.readout_priority[scan_id] = {}
    with mock.patch.object(sb, "run_emitter") as emitter:
        sb._baseline_update(scan_id, device, signal)
        emitter.assert_called_once_with("on_baseline_emit", scan_id)


def test_update_monitor_signals(scan_bundler_mock):
    scan_id = "ljlaskdj"
    point_id = 2
    sb = scan_bundler_mock
    sb.sync_storage[scan_id] = {"info": {"scan_type": "fly"}, point_id: {}}
    sb.monitored_devices[scan_id] = {
        "devices": sb.device_manager.devices.monitored_devices([]),
        "point_id": {},
    }
    num_devices = len(sb.device_manager.devices.monitored_devices([]))
    with mock.patch.object(
        sb, "_get_last_device_readback", return_value=[{"value": 400} for _ in range(num_devices)]
    ):
        sb._update_monitor_signals(scan_id, point_id)
        assert sb.sync_storage[scan_id][point_id]["bpm3a"] == {"value": 400}


def test_get_last_device_readback(scan_bundler_mock):
    sb = scan_bundler_mock
    dev_msg = messages.DeviceMessage(
        signals={
            "samx": {"value": 0.51},
            "setpoint": {"value": 0.5},
            "motor_is_moving": {"value": 0},
        },
        metadata={"scan_id": "laksjd", "readout_priority": "monitored"},
    )
    with mock.patch.object(sb, "connector") as connector_mock:
        connector_mock.execute_pipeline.return_value = [dev_msg]
        ret = sb._get_last_device_readback([sb.device_manager.devices.samx])
        assert connector_mock.get.mock_calls == [
            mock.call(MessageEndpoints.device_readback("samx"), connector_mock.pipeline())
        ]
        assert ret == [dev_msg.content["signals"]]



"""
This module contains the CallbackHandler class to handle callbacks. 
The CallbackHandler class is used to register and run callbacks for different 
event types. The CallbackRegister class is used to register callbacks 
in a with statement.
"""

import builtins
import enum
import threading
import traceback
from collections import deque
from collections.abc import Callable

from bec_lib.logger import bec_logger
from bec_lib.utils import threadlocked

logger = bec_logger.logger


class EventType(str, enum.Enum):
    """Event types"""

    SCAN_SEGMENT = "scan_segment"
    SCAN_STATUS = "scan_status"
    NAMESPACE_UPDATE = "namespace_update"


class CallbackEntry:
    """Callback entry class to store callback information"""

    def __init__(self, id: int, event_type: EventType, func: Callable, sync: bool) -> None:
        self.id = id
        self.func = func
        self.event_type = event_type
        self.sync = sync
        self.queue = deque(maxlen=1000)
        self._lock = threading.RLock()

    @threadlocked
    def run(self, *args, **kwargs) -> None:
        """Run the callback function. If sync is True, the callback is run immediately. Otherwise, the callback is added to a queue and exectued in the next poll."""
        if not self.sync:
            self._run_cb(*args, **kwargs)
            return
        self.queue.append((args, kwargs))

    def _run_cb(self, *args, **kwargs) -> None:
        """Run the callback function in a safe way."""
        try:
            self.func(*args, **kwargs)
        except Exception:
            content = traceback.format_exc()
            logger.warning(f"Failed to run callback function: {content}")

    def __str__(self) -> str:
        return f"<CallbackEntry>: (event_type: {self.event_type}, function: {self.func.__name__}, sync: {self.sync}, pending events: {self.num_pending_events})"

    @property
    def num_pending_events(self):
        """number of pending events"""
        return len(self.queue)

    @threadlocked
    def poll(self) -> None:
        """Run callback.

        Raises:
            RuntimeError: Raises if attempt is made to run async callbacks manually.
        """
        if not self.sync:
            raise RuntimeError("Cannot poll on an async callback.")
        args, kwargs = self.queue.popleft()
        self._run_cb(*args, **kwargs)


class CallbackHandler:
    """Callback handler class"""

    def __init__(self) -> None:
        self.callbacks = {}
        self.id_counter = 0
        self._lock = threading.RLock()

    @threadlocked
    def register(self, event_type: str, callback: Callable, sync=False) -> int:
        """Register a callback to an event type

        Args:
            event_type (str): Event type
            callback (Callable): Callback function
            sync (bool, optional): Synchronous or async callback. Defaults to False.

        Returns:
            int: Callback id
        """
        event_type = EventType(event_type)
        callback_id = self.new_id()
        self.callbacks[callback_id] = CallbackEntry(callback_id, event_type, callback, sync)
        return callback_id

    @threadlocked
    def register_many(self, event_type: str, callbacks: list[Callable], sync=False) -> list[int]:
        """Register multiple callbacks to an event type

        Args:
            event_type (str): Event type
            callbacks (list[Callable]): List of callback functions
            sync (bool, optional): Synchronous or async callback. Defaults to False.

        Returns:
            list: List of caallback ids
        """

        if not isinstance(callbacks, list):
            callbacks = [callbacks]
        ids = []
        for cbk in callbacks:
            if cbk:
                ids.append(self.register(event_type, cbk, sync))
            else:
                ids.append(-1)
        return ids

    @threadlocked
    def remove(self, id: int) -> int:
        """Remove a registered callback by its id

        Args:
            id (int): Callback id

        Returns:
            int: Returns the id of the removed callback. -1 if it failed.
        """
        try:
            self.callbacks.pop(id)
            return id
        except KeyError:
            return -1

    def new_id(self):
        """Generate a new callback id"""
        self.id_counter += 1
        return self.id_counter

    @threadlocked
    def run(self, event_type: str, *args, **kwargs):
        """Run all callbacks for a given event type"""
        for cb in self.callbacks.values():
            if event_type != cb.event_type:
                continue
            cb.run(*args, **kwargs)

    @threadlocked
    def poll(self):
        """Run all pending callbacks"""
        for callback in self.callbacks.values():
            if not callback.sync:
                continue
            while callback.num_pending_events:
                callback.poll()


class CallbackRegister:
    def __init__(self, event_type, callbacks, sync=False, callback_handler=None) -> None:
        """Callback register class to register callbacks in a with statement

        Args:
            callback_handler (CallbackHandler): Callback handler
        """
        if not callback_handler:
            bec = builtins.__dict__.get("bec")
            self.callback_handler = bec.callbacks
        else:
            self.callback_handler = callback_handler
        self.event_type = event_type
        if not isinstance(callbacks, list):
            callbacks = [callbacks]
        self.callbacks = callbacks
        self.sync = sync
        self.callback_ids = []

    def __enter__(self):
        for callback in self.callbacks:
            if not callback:
                continue
            self.callback_ids.append(
                self.callback_handler.register(self.event_type, callback, sync=self.sync)
            )
        return self

    def __exit__(self, *exc):
        for cb_id in self.callback_ids:
            self.callback_handler.remove(cb_id)



from ophyd import EpicsMotor, EpicsSignal, EpicsSignalRO
from ophyd.quadem import QuadEM
from ophyd.sim import SynAxis, SynPeriodicSignal, SynSignal

from .epics_motor_ex import EpicsMotorEx
from .slits import SlitH, SlitV
from .sls_devices import SLSInfo, SLSOperatorMessages
from .specMotors import (
    Bpm4i,
    EnergyKev,
    GirderMotorPITCH,
    GirderMotorROLL,
    GirderMotorX1,
    GirderMotorY1,
    GirderMotorYAW,
    MonoTheta1,
    MonoTheta2,
    PmDetectorRotation,
    PmMonoBender,
)
from .SpmBase import SpmBase



import os


def setup_epics_ca():
    # os.environ["EPICS_CA_AUTO_ADDR_LIST"] = "NO"
    # os.environ["EPICS_CA_ADDR_LIST"] = "129.129.122.255 sls-x12sa-cagw.psi.ch:5836"
    os.environ["PYTHONIOENCODING"] = "latin1"


def run():
    setup_epics_ca()



from bec_lib.device import DeviceBase
from bec_lib.utils.import_utils import isinstance_based_on_class_name


def test_isinstance_based_on_class_name():
    obj = DeviceBase(name="test_obj")

    assert isinstance_based_on_class_name(obj, "bec_lib.device.DeviceBase")
    assert not isinstance_based_on_class_name(obj, "bec_lib.device.Status")
